from __future__ import unicode_literals

from collections import defaultdict
import json
import logging

from django.conf import settings
from django.contrib import auth
from django.core import exceptions as django_exceptions
from django.core.validators import RegexValidator, MaxLengthValidator
from django.db import models as django_models, transaction
from django.db.models import Q
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
import pyvat
from rest_framework import exceptions, serializers
from rest_framework.reverse import reverse
import six

from waldur_core.core import (models as core_models, fields as core_fields, serializers as core_serializers,
                              utils as core_utils)
from waldur_core.core.fields import MappedChoiceField
from waldur_core.monitoring.serializers import MonitoringSerializerMixin
from waldur_core.quotas import serializers as quotas_serializers
from waldur_core.structure import (models, SupportedServices, ServiceBackendError, ServiceBackendNotImplemented,
                                   executors)
from waldur_core.structure.filters import filter_visible_users
from waldur_core.structure.managers import filter_queryset_for_user

User = auth.get_user_model()
logger = logging.getLogger(__name__)


class IpCountValidator(MaxLengthValidator):
    message = _('Only %(limit_value)s ip address is supported.')


class PermissionFieldFilteringMixin(object):
    """
    Mixin allowing to filter related fields.

    In order to constrain the list of entities that can be used
    as a value for the field:

    1. Make sure that the entity in question has corresponding
       Permission class defined.

    2. Implement `get_filtered_field_names()` method
       in the class that this mixin is mixed into and return
       the field in question from that method.
    """

    def get_fields(self):
        fields = super(PermissionFieldFilteringMixin, self).get_fields()

        try:
            request = self.context['request']
            user = request.user
        except (KeyError, AttributeError):
            return fields

        for field_name in self.get_filtered_field_names():
            if field_name not in fields:  # field could be not required by user
                continue
            field = fields[field_name]
            field.queryset = filter_queryset_for_user(field.queryset, user)

        return fields

    def get_filtered_field_names(self):
        raise NotImplementedError(
            'Implement get_filtered_field_names() '
            'to return list of filtered fields')


class PermissionListSerializer(serializers.ListSerializer):
    """
    Allows to filter related queryset by user.
    Counterpart of PermissionFieldFilteringMixin.

    In order to use it set Meta.list_serializer_class. Example:

    >>> class PermissionProjectSerializer(BasicProjectSerializer):
    >>>     class Meta(BasicProjectSerializer.Meta):
    >>>         list_serializer_class = PermissionListSerializer
    >>>
    >>> class CustomerSerializer(serializers.HyperlinkedModelSerializer):
    >>>     projects = PermissionProjectSerializer(many=True, read_only=True)
    """

    def to_representation(self, data):
        try:
            request = self.context['request']
            user = request.user
        except (KeyError, AttributeError):
            pass
        else:
            if isinstance(data, (django_models.Manager, django_models.query.QuerySet)):
                data = filter_queryset_for_user(data.all(), user)

        return super(PermissionListSerializer, self).to_representation(data)


class BasicUserSerializer(serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = User
        fields = ('url', 'uuid', 'username', 'full_name', 'native_name',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }


class BasicProjectSerializer(core_serializers.BasicInfoSerializer):
    class Meta(core_serializers.BasicInfoSerializer.Meta):
        model = models.Project


class PermissionProjectSerializer(BasicProjectSerializer):
    class Meta(BasicProjectSerializer.Meta):
        list_serializer_class = PermissionListSerializer


class NestedServiceProjectLinkSerializer(serializers.Serializer):
    uuid = serializers.ReadOnlyField(source='service.uuid')
    url = serializers.SerializerMethodField()
    service_project_link_url = serializers.SerializerMethodField()
    name = serializers.ReadOnlyField(source='service.settings.name')
    type = serializers.SerializerMethodField()
    state = serializers.SerializerMethodField()
    shared = serializers.SerializerMethodField()
    settings_uuid = serializers.ReadOnlyField(source='service.settings.uuid')
    settings = serializers.SerializerMethodField()
    validation_state = serializers.ChoiceField(
        choices=models.ServiceProjectLink.States.CHOICES,
        read_only=True,
        help_text=_('A state of service compliance with project requirements.'))
    validation_message = serializers.ReadOnlyField(
        help_text=_('An error message for a service that is non-compliant with project requirements.'))

    def get_settings(self, link):
        """
        URL of service settings
        """
        return reverse(
            'servicesettings-detail', kwargs={'uuid': link.service.settings.uuid}, request=self.context['request'])

    def get_url(self, link):
        """
        URL of service
        """
        view_name = SupportedServices.get_detail_view_for_model(link.service)
        return reverse(view_name, kwargs={'uuid': link.service.uuid.hex}, request=self.context['request'])

    def get_service_project_link_url(self, link):
        view_name = SupportedServices.get_detail_view_for_model(link)
        return reverse(view_name, kwargs={'pk': link.id}, request=self.context['request'])

    def get_type(self, link):
        return SupportedServices.get_name_for_model(link.service)

    # XXX: SPL is intended to become stateless. For backward compatiblity we are returning here state from connected
    # service settings. To be removed once SPL becomes stateless.
    def get_state(self, link):
        return link.service.settings.get_state_display()

    def get_resources_count(self, link):
        """
        Count total number of all resources connected to link
        """
        total = 0
        for model in SupportedServices.get_service_resources(link.service):
            # Format query path from resource to service project link
            query = {model.Permissions.project_path.split('__')[0]: link}
            total += model.objects.filter(**query).count()
        return total

    def get_shared(self, link):
        return link.service.settings.shared


class NestedServiceCertificationSerializer(core_serializers.AugmentedSerializerMixin,
                                           core_serializers.HyperlinkedRelatedModelSerializer):
    class Meta(object):
        model = models.ServiceCertification
        fields = ('uuid', 'url', 'name', 'description', 'link')
        read_only_fields = ('name', 'description', 'link')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }


class ProjectTypeSerializer(serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = models.ProjectType
        fields = ('uuid', 'url', 'name', 'description')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'project_type-detail'},
        }


class ProjectSerializer(core_serializers.RestrictedSerializerMixin,
                        PermissionFieldFilteringMixin,
                        core_serializers.AugmentedSerializerMixin,
                        serializers.HyperlinkedModelSerializer):
    quotas = quotas_serializers.BasicQuotaSerializer(many=True, read_only=True)
    services = serializers.SerializerMethodField()
    certifications = NestedServiceCertificationSerializer(
        queryset=models.ServiceCertification.objects.all(),
        many=True, required=False)

    class Meta(object):
        model = models.Project
        fields = (
            'url', 'uuid',
            'name',
            'customer', 'customer_uuid', 'customer_name', 'customer_native_name', 'customer_abbreviation',
            'description',
            'quotas',
            'services',
            'created',
            'certifications',
            'type', 'type_name',
        )
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'customer': {'lookup_field': 'uuid'},
            'certifications': {'lookup_field': 'uuid'},
            'type': {'lookup_field': 'uuid', 'view_name': 'project_type-detail'},
        }
        related_paths = {
            'customer': ('uuid', 'name', 'native_name', 'abbreviation'),
            'type': ('name',),
        }
        protected_fields = ('certifications',)

    @staticmethod
    def eager_load(queryset, request=None):
        related_fields = (
            'uuid',
            'name',
            'created',
            'description',
            'customer__uuid',
            'customer__name',
            'customer__native_name',
            'customer__abbreviation',
        )
        return queryset.select_related('customer').only(*related_fields) \
            .prefetch_related('quotas', 'certifications')

    def create(self, validated_data):
        certifications = validated_data.pop('certifications', [])
        project = super(ProjectSerializer, self).create(validated_data)
        project.certifications.add(*certifications)

        return project

    def get_filtered_field_names(self):
        return 'customer',

    def get_services(self, project):
        if 'services' not in self.context:
            self.context['services'] = self.get_services_map()
        services = self.context['services'][project.pk]

        serializer = NestedServiceProjectLinkSerializer(
            services,
            many=True,
            read_only=True,
            context={'request': self.context['request']})
        return serializer.data

    def get_services_map(self):
        services = defaultdict(list)
        related_fields = (
            'id',
            'service__settings__state',
            'project_id',
            'service__uuid',
            'service__settings__uuid',
            'service__settings__shared',
            'service__settings__name',
        )
        for link_model in models.ServiceProjectLink.get_all_models():
            links = (link_model.objects.all()
                     .select_related('service', 'service__settings')
                     .only(*related_fields)
                     .prefetch_related('service__settings__certifications'))
            if isinstance(self.instance, list):
                links = links.filter(project__in=self.instance)
            else:
                links = links.filter(project=self.instance)
            for link in links:
                services[link.project_id].append(link)
        return services


class CustomerSerializer(core_serializers.RestrictedSerializerMixin,
                         core_serializers.AugmentedSerializerMixin,
                         serializers.HyperlinkedModelSerializer, ):
    projects = PermissionProjectSerializer(many=True, read_only=True)
    owners = BasicUserSerializer(source='get_owners', many=True, read_only=True)
    support_users = BasicUserSerializer(source='get_support_users', many=True, read_only=True)
    quotas = quotas_serializers.BasicQuotaSerializer(many=True, read_only=True)

    COUNTRIES = core_fields.CountryField.COUNTRIES
    if settings.WALDUR_CORE.get('COUNTRIES'):
        COUNTRIES = [item for item in COUNTRIES if item[0] in settings.WALDUR_CORE['COUNTRIES']]
    country = serializers.ChoiceField(required=False, choices=COUNTRIES, allow_blank=True)
    country_name = serializers.ReadOnlyField(source='get_country_display')
    display_name = serializers.ReadOnlyField(source='get_display_name')
    division_name = serializers.ReadOnlyField(source='division.name')
    division_uuid = serializers.ReadOnlyField(source='division.uuid')

    class Meta(object):
        model = models.Customer
        fields = (
            'url',
            'uuid',
            'created',
            'name', 'native_name', 'abbreviation',
            'division', 'division_name', 'division_uuid',
            'contact_details',
            'domain', 'display_name',
            'agreement_number', 'email', 'phone_number', 'access_subnets',
            'projects',
            'owners', 'support_users',
            'backend_id', 'registration_code', 'homepage',
            'quotas',
            'image',
            'country', 'country_name', 'vat_code', 'is_company',
            'type', 'postal', 'address', 'bank_name', 'bank_account',
            'default_tax_percent', 'accounting_start_date',
        )
        protected_fields = ('agreement_number',)
        read_only_fields = ('access_subnets', 'accounting_start_date', 'default_tax_percent')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'division': {'lookup_field': 'uuid'},
        }

    def get_fields(self):
        fields = super(CustomerSerializer, self).get_fields()

        try:
            request = self.context['view'].request
            user = request.user
        except (KeyError, AttributeError):
            return fields

        if not user.is_staff:
            if 'domain' in fields:
                fields['domain'].read_only = True

        return fields

    def create(self, validated_data):
        user = self.context['request'].user
        if 'domain' not in validated_data:
            # Staff can specify domain name on organization creation
            validated_data['domain'] = user.organization
        return super(CustomerSerializer, self).create(validated_data)

    @staticmethod
    def eager_load(queryset, request=None):
        return queryset.prefetch_related('quotas', 'projects')

    def validate(self, attrs):
        country = attrs.get('country')
        vat_code = attrs.get('vat_code')
        is_company = attrs.get('is_company')

        if vat_code:
            if not is_company:
                raise serializers.ValidationError({
                    'vat_code': _('VAT number is not supported for private persons.')})

            # Check VAT format
            if not pyvat.is_vat_number_format_valid(vat_code, country):
                raise serializers.ValidationError({'vat_code': _('VAT number has invalid format.')})

            # Check VAT number in EU VAT Information Exchange System
            # if customer is new or either VAT number or country of the customer has changed
            if not self.instance or self.instance.vat_code != vat_code or self.instance.country != country:
                check_result = pyvat.check_vat_number(vat_code, country)
                if check_result.is_valid:
                    attrs['vat_name'] = check_result.business_name
                    attrs['vat_address'] = check_result.business_address
                    if not attrs.get('contact_details'):
                        attrs['contact_details'] = attrs['vat_address']
                elif check_result.is_valid is False:
                    raise serializers.ValidationError({'vat_code': _('VAT number is invalid.')})
                else:
                    logger.debug('Unable to check VAT number %s for country %s. Error message: %s',
                                 vat_code, country, check_result.log_lines)
                    raise serializers.ValidationError({'vat_code': _('Unable to check VAT number.')})
        return attrs


class NestedProjectPermissionSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedRelatedField(
        source='project',
        lookup_field='uuid',
        view_name='project-detail',
        queryset=models.Project.objects.all(),
    )
    uuid = serializers.ReadOnlyField(source='project.uuid')
    name = serializers.ReadOnlyField(source='project.name')
    permission = serializers.HyperlinkedRelatedField(
        source='pk',
        view_name='project_permission-detail',
        queryset=models.ProjectPermission.objects.all(),
    )

    class Meta:
        model = models.ProjectPermission
        fields = ['url', 'uuid', 'name', 'role', 'permission', 'expiration_time']


class CustomerUserSerializer(serializers.ModelSerializer):
    role = serializers.ReadOnlyField()
    expiration_time = serializers.ReadOnlyField(source='perm.expiration_time')
    permission = serializers.HyperlinkedRelatedField(
        source='perm.pk',
        view_name='customer_permission-detail',
        queryset=models.CustomerPermission.objects.all(),
    )
    projects = NestedProjectPermissionSerializer(many=True, read_only=True)

    class Meta:
        model = User
        fields = ['url', 'uuid', 'username', 'full_name', 'email', 'role', 'permission', 'projects',
                  'expiration_time']
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    def to_representation(self, user):
        customer = self.context['customer']
        permission = models.CustomerPermission.objects.filter(
            customer=customer, user=user, is_active=True).first()
        projects = models.ProjectPermission.objects.filter(
            project__customer=customer, user=user, is_active=True)
        setattr(user, 'perm', permission)
        setattr(user, 'role', permission and permission.role)
        setattr(user, 'projects', projects)
        return super(CustomerUserSerializer, self).to_representation(user)


class ProjectUserSerializer(serializers.ModelSerializer):
    role = serializers.ReadOnlyField()
    expiration_time = serializers.ReadOnlyField(source='perm.expiration_time')
    permission = serializers.HyperlinkedRelatedField(
        source='perm.pk',
        view_name='project_permission-detail',
        queryset=models.ProjectPermission.objects.all(),
    )

    class Meta:
        model = User
        fields = ['url', 'uuid', 'username', 'full_name', 'email', 'role', 'permission',
                  'expiration_time']
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    def to_representation(self, user):
        project = self.context['project']
        permission = models.ProjectPermission.objects.filter(
            project=project, user=user, is_active=True).first()
        setattr(user, 'perm', permission)
        setattr(user, 'role', permission and permission.role)
        return super(ProjectUserSerializer, self).to_representation(user)


class BasePermissionSerializer(core_serializers.AugmentedSerializerMixin, serializers.HyperlinkedModelSerializer):
    class Meta(object):
        fields = ('user', 'user_full_name', 'user_native_name', 'user_username', 'user_uuid', 'user_email')
        related_paths = {
            'user': ('username', 'full_name', 'native_name', 'uuid', 'email'),
        }


class BasicCustomerPermissionSerializer(BasePermissionSerializer):
    class Meta(BasePermissionSerializer.Meta):
        model = models.CustomerPermission
        fields = (
            'url', 'pk', 'role', 'customer_uuid', 'customer_name', 'customer_native_name', 'customer_abbreviation',
        )
        related_paths = dict(
            customer=('name', 'native_name', 'abbreviation', 'uuid'),
            **BasePermissionSerializer.Meta.related_paths
        )
        extra_kwargs = {
            'customer': {
                'view_name': 'customer-detail',
                'lookup_field': 'uuid',
                'queryset': models.Customer.objects.all(),
            }
        }


class CustomerPermissionSerializer(PermissionFieldFilteringMixin, BasePermissionSerializer):
    class Meta(BasePermissionSerializer.Meta):
        model = models.CustomerPermission
        fields = (
            'url', 'pk', 'role', 'created', 'expiration_time', 'created_by',
            'customer', 'customer_uuid', 'customer_name', 'customer_native_name', 'customer_abbreviation',
        ) + BasePermissionSerializer.Meta.fields
        related_paths = dict(
            customer=('name', 'native_name', 'abbreviation', 'uuid'),
            **BasePermissionSerializer.Meta.related_paths
        )
        protected_fields = (
            'customer', 'role', 'user', 'created_by', 'created'
        )
        extra_kwargs = {
            'user': {
                'view_name': 'user-detail',
                'lookup_field': 'uuid',
                'queryset': User.objects.all(),
            },
            'created_by': {
                'view_name': 'user-detail',
                'lookup_field': 'uuid',
                'read_only': True,
            },
            'customer': {
                'view_name': 'customer-detail',
                'lookup_field': 'uuid',
                'queryset': models.Customer.objects.all(),
            }
        }

    def validate(self, data):
        if not self.instance:
            customer = data['customer']
            user = data['user']

            if customer.has_user(user):
                raise serializers.ValidationError(_('The fields customer and user must make a unique set.'))

        return data

    def create(self, validated_data):
        customer = validated_data['customer']
        user = validated_data['user']
        role = validated_data['role']
        expiration_time = validated_data.get('expiration_time')

        created_by = self.context['request'].user
        permission, _ = customer.add_user(user, role, created_by, expiration_time)

        return permission

    def validate_expiration_time(self, value):
        if value is not None and value < timezone.now():
            raise serializers.ValidationError(_('Expiration time should be greater than current time.'))
        return value

    def get_filtered_field_names(self):
        return ('customer',)


class CustomerPermissionLogSerializer(CustomerPermissionSerializer):
    class Meta(CustomerPermissionSerializer.Meta):
        view_name = 'customer_permission_log-detail'


class ProjectPermissionSerializer(PermissionFieldFilteringMixin, BasePermissionSerializer):
    customer_name = serializers.ReadOnlyField(source='project.customer.name')

    class Meta(BasePermissionSerializer.Meta):
        model = models.ProjectPermission
        fields = (
            'url', 'pk', 'role', 'created', 'expiration_time', 'created_by',
            'project', 'project_uuid', 'project_name', 'customer_name'
        ) + BasePermissionSerializer.Meta.fields
        related_paths = dict(
            project=('name', 'uuid'),
            **BasePermissionSerializer.Meta.related_paths
        )
        protected_fields = (
            'project', 'role', 'user', 'created_by', 'created'
        )
        extra_kwargs = {
            'user': {
                'view_name': 'user-detail',
                'lookup_field': 'uuid',
                'queryset': User.objects.all(),
            },
            'created_by': {
                'view_name': 'user-detail',
                'lookup_field': 'uuid',
                'read_only': True,
            },
            'project': {
                'view_name': 'project-detail',
                'lookup_field': 'uuid',
                'queryset': models.Project.objects.all(),
            }
        }

    def validate(self, data):
        if not self.instance:
            project = data['project']
            user = data['user']

            if project.has_user(user):
                raise serializers.ValidationError(_('The fields project and user must make a unique set.'))

        return data

    def create(self, validated_data):
        project = validated_data['project']
        user = validated_data['user']
        role = validated_data['role']
        expiration_time = validated_data.get('expiration_time')

        created_by = self.context['request'].user
        permission, _ = project.add_user(user, role, created_by, expiration_time)

        return permission

    def validate_expiration_time(self, value):
        if value is not None and value < timezone.now():
            raise serializers.ValidationError(_('Expiration time should be greater than current time.'))
        return value

    def get_filtered_field_names(self):
        return ('project',)


class BasicProjectPermissionSerializer(BasePermissionSerializer):
    customer_name = serializers.ReadOnlyField(source='project.customer.name')

    class Meta(BasePermissionSerializer.Meta):
        model = models.ProjectPermission
        fields = (
            'url', 'pk', 'role', 'project_uuid', 'project_name', 'customer_name',
        )
        related_paths = dict(
            project=('name', 'uuid'),
            **BasePermissionSerializer.Meta.related_paths
        )
        extra_kwargs = {
            'project': {
                'view_name': 'project-detail',
                'lookup_field': 'uuid',
                'queryset': models.Project.objects.all(),
            }
        }


class ProjectPermissionLogSerializer(ProjectPermissionSerializer):
    class Meta(ProjectPermissionSerializer.Meta):
        view_name = 'project_permission_log-detail'


class UserSerializer(serializers.HyperlinkedModelSerializer):
    email = serializers.EmailField()
    agree_with_policy = serializers.BooleanField(write_only=True, required=False,
                                                 help_text=_('User must agree with the policy to register.'))
    competence = serializers.ChoiceField(choices=settings.WALDUR_CORE.get('USER_COMPETENCE_LIST', []),
                                         allow_blank=True,
                                         required=False)
    token = serializers.ReadOnlyField(source='auth_token.key')
    customer_permissions = serializers.SerializerMethodField()
    project_permissions = serializers.SerializerMethodField()

    def get_customer_permissions(self, user):
        permissions = models.CustomerPermission.objects.filter(user=user, is_active=True).select_related('customer')
        serializer = BasicCustomerPermissionSerializer(instance=permissions, many=True,
                                                       context=self.context)
        return serializer.data

    def get_project_permissions(self, user):
        permissions = models.ProjectPermission.objects.filter(user=user, is_active=True).select_related('project')
        serializer = BasicProjectPermissionSerializer(instance=permissions, many=True,
                                                      context=self.context)
        return serializer.data

    class Meta(object):
        model = User
        fields = (
            'url',
            'uuid', 'username',
            'full_name', 'native_name',
            'job_title', 'email', 'phone_number',
            'organization',
            'civil_number',
            'description',
            'is_staff', 'is_active', 'is_support',
            'token', 'token_lifetime',
            'registration_method',
            'date_joined',
            'agree_with_policy',
            'agreement_date',
            'preferred_language',
            'competence',
            'customer_permissions',
            'project_permissions',
        )
        read_only_fields = (
            'uuid',
            'civil_number',
            'registration_method',
            'date_joined',
            'agreement_date',
            'customer_permissions',
            'project_permissions',
        )
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    def get_fields(self):
        fields = super(UserSerializer, self).get_fields()

        try:
            request = self.context['view'].request
            user = request.user
        except (KeyError, AttributeError):
            return fields

        if not user.is_staff:
            protected_fields = ('is_active', 'is_staff', 'is_support', 'description')
            if user.is_support:
                for field in protected_fields:
                    if field in fields:
                        fields[field].read_only = True
            else:
                for field in protected_fields:
                    if field in fields:
                        del fields[field]

        if not self._can_see_token(user):
            del fields['token']
            del fields['token_lifetime']

        if request.method in ('PUT', 'PATCH'):
            fields['username'].read_only = True
            protected_methods = settings.WALDUR_CORE['PROTECT_USER_DETAILS_FOR_REGISTRATION_METHODS']
            if user.registration_method and user.registration_method in protected_methods:
                detail_fields = (
                    'full_name', 'native_name',
                    'job_title', 'email', 'phone_number',
                    'organization',
                )
                for field in detail_fields:
                    fields[field].read_only = True

        return fields

    def _can_see_token(self, user):
        # Staff can see any token
        # User can see his own token either via details view or /api/users/?current

        if user.is_staff:
            return True
        elif isinstance(self.instance, list) and len(self.instance) == 1:
            return self.instance[0] == user
        else:
            return self.instance == user

    def validate(self, attrs):
        agree_with_policy = attrs.pop('agree_with_policy', False)
        if self.instance and not self.instance.agreement_date:
            if not agree_with_policy:
                raise serializers.ValidationError({'agree_with_policy': _('User must agree with the policy.')})
            else:
                attrs['agreement_date'] = timezone.now()

        # Convert validation error from Django to DRF
        # https://github.com/tomchristie/django-rest-framework/issues/2145
        try:
            user = User(id=getattr(self.instance, 'id', None), **attrs)
            user.clean()
        except django_exceptions.ValidationError as error:
            raise exceptions.ValidationError(error.message_dict)
        return attrs


class CreationTimeStatsSerializer(serializers.Serializer):
    MODEL_NAME_CHOICES = (('project', 'project'), ('customer', 'customer'),)
    MODEL_CLASSES = {'project': models.Project, 'customer': models.Customer}

    model_name = serializers.ChoiceField(choices=MODEL_NAME_CHOICES)
    start_timestamp = serializers.IntegerField(min_value=0)
    end_timestamp = serializers.IntegerField(min_value=0)
    segments_count = serializers.IntegerField(min_value=0)

    def get_stats(self, user):
        start_datetime = core_utils.timestamp_to_datetime(self.data['start_timestamp'])
        end_datetime = core_utils.timestamp_to_datetime(self.data['end_timestamp'])

        model = self.MODEL_CLASSES[self.data['model_name']]
        filtered_queryset = filter_queryset_for_user(model.objects.all(), user)
        created_datetimes = (
            filtered_queryset
            .filter(created__gte=start_datetime, created__lte=end_datetime)
            .values('created')
            .annotate(count=django_models.Count('id', distinct=True)))

        time_and_value_list = [
            (core_utils.datetime_to_timestamp(dt['created']), dt['count']) for dt in created_datetimes]

        return core_utils.format_time_and_value_to_segment_list(
            time_and_value_list, self.data['segments_count'],
            self.data['start_timestamp'], self.data['end_timestamp'])


class PasswordSerializer(serializers.Serializer):
    password = serializers.CharField(min_length=7, validators=[
        RegexValidator(
            regex='\d',
            message=_('Ensure this field has at least one digit.'),
        ),
        RegexValidator(
            regex='[a-zA-Z]',
            message=_('Ensure this field has at least one latin letter.'),
        ),
    ])


class SshKeySerializer(serializers.HyperlinkedModelSerializer):
    user_uuid = serializers.ReadOnlyField(source='user.uuid')

    class Meta(object):
        model = core_models.SshPublicKey
        fields = ('url', 'uuid', 'name', 'public_key', 'fingerprint', 'user_uuid', 'is_shared')
        read_only_fields = ('fingerprint', 'is_shared')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    def validate_name(self, value):
        return value.strip()

    def validate_public_key(self, value):
        value = value.strip()
        if len(value.splitlines()) > 1:
            raise serializers.ValidationError(_('Key is not valid: it should be single line.'))

        try:
            fingerprint = core_models.get_ssh_key_fingerprint(value)
        except (IndexError, TypeError):
            raise serializers.ValidationError(_('Key is not valid: cannot generate fingerprint from it.'))
        if core_models.SshPublicKey.objects.filter(fingerprint=fingerprint).exists():
            raise serializers.ValidationError(_('Key with same fingerprint already exists.'))
        return value

    def get_fields(self):
        fields = super(SshKeySerializer, self).get_fields()

        try:
            user = self.context['request'].user
        except (KeyError, AttributeError):
            return fields

        if not user.is_staff:
            del fields['user_uuid']

        return fields


class ServiceCertificationsUpdateSerializer(serializers.Serializer):
    certifications = NestedServiceCertificationSerializer(
        queryset=models.ServiceCertification.objects.all(),
        required=True,
        many=True)

    @transaction.atomic
    def update(self, instance, validated_data):
        certifications = validated_data.pop('certifications', None)
        instance.certifications.clear()
        instance.certifications.add(*certifications)
        return instance


class ServiceCertificationSerializer(serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = models.ServiceCertification
        fields = ('uuid', 'url', 'name', 'description', 'link')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'service-certification-detail'},
        }


class ServiceSettingsSerializer(PermissionFieldFilteringMixin,
                                core_serializers.RestrictedSerializerMixin,
                                core_serializers.AugmentedSerializerMixin,
                                serializers.HyperlinkedModelSerializer):
    customer_native_name = serializers.ReadOnlyField(source='customer.native_name')
    state = MappedChoiceField(
        choices=[(v, k) for k, v in core_models.StateMixin.States.CHOICES],
        choice_mappings={v: k for k, v in core_models.StateMixin.States.CHOICES},
        read_only=True)
    quotas = quotas_serializers.BasicQuotaSerializer(many=True, read_only=True)
    scope = core_serializers.GenericRelatedField(related_models=models.ResourceMixin.get_all_models(), required=False)
    certifications = NestedServiceCertificationSerializer(many=True, read_only=True)
    geolocations = core_serializers.GeoLocationField(read_only=True)

    class Meta(object):
        model = models.ServiceSettings
        fields = (
            'url', 'uuid', 'name', 'type', 'state', 'error_message', 'shared',
            'backend_url', 'username', 'password', 'token', 'certificate',
            'customer', 'customer_name', 'customer_native_name',
            'homepage', 'terms_of_services', 'certifications',
            'quotas', 'scope', 'geolocations',
        )
        protected_fields = ('type', 'customer')
        read_only_fields = ('shared', 'state', 'error_message')
        related_paths = ('customer',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'customer': {'lookup_field': 'uuid'},
            'certifications': {'lookup_field': 'uuid'},
        }
        write_only_fields = ('backend_url', 'username', 'token', 'password', 'certificate')
        for field in write_only_fields:
            field_params = extra_kwargs.setdefault(field, {})
            field_params['write_only'] = True

    def get_filtered_field_names(self):
        return 'customer',

    @staticmethod
    def eager_load(queryset, request=None):
        return queryset.select_related('customer').prefetch_related('quotas', 'certifications')

    def get_fields(self):
        fields = super(ServiceSettingsSerializer, self).get_fields()
        request = self.context['request']

        if isinstance(self.instance, self.Meta.model):
            if self.can_see_extra_fields():
                # If user can change settings he should be able to see value
                for field in self.Meta.write_only_fields:
                    fields[field].write_only = False

                serializer = self.get_service_serializer()

                # Remove fields if they are not needed for service
                filter_fields = serializer.SERVICE_ACCOUNT_FIELDS
                if filter_fields is not NotImplemented:
                    for field in self.Meta.write_only_fields:
                        if field in filter_fields:
                            fields[field].help_text = filter_fields[field]
                        elif field in fields:
                            del fields[field]

                # Add extra fields stored in options dictionary
                extra_fields = serializer.SERVICE_ACCOUNT_EXTRA_FIELDS
                if extra_fields is not NotImplemented:
                    for field in extra_fields:
                        fields[field] = serializers.CharField(required=False,
                                                              source='options.' + field,
                                                              allow_blank=True,
                                                              help_text=extra_fields[field])

        if request.method == 'GET':
            fields['type'] = serializers.ReadOnlyField(source='get_type_display')

        return fields

    def get_service_serializer(self):
        service = SupportedServices.get_service_models()[self.instance.type]['service']
        # Find service serializer by service type of settings object
        return next(cls for cls in BaseServiceSerializer.__subclasses__()
                    if cls.Meta.model == service)

    def can_see_extra_fields(self):
        request = self.context['request']

        if request.user.is_staff:
            return True

        if not self.instance.customer:
            return False

        return self.instance.customer.has_user(request.user, models.CustomerRole.OWNER)

    def update(self, instance, validated_data):
        if 'options' in validated_data:
            new_options = dict.copy(instance.options)
            new_options.update(validated_data['options'])
            validated_data['options'] = new_options

        return super(ServiceSettingsSerializer, self).update(instance, validated_data)


class ServiceSerializerMetaclass(serializers.SerializerMetaclass):
    """ Build a list of supported services via serializers definition.
        See SupportedServices for details.
    """

    def __new__(cls, name, bases, args):
        SupportedServices.register_service(args['Meta'].model)
        serializer = super(ServiceSerializerMetaclass, cls).__new__(cls, name, bases, args)
        SupportedServices.register_service_serializer(args['Meta'].model, serializer)
        return serializer


class BaseServiceSerializer(six.with_metaclass(ServiceSerializerMetaclass,
                                               PermissionFieldFilteringMixin,
                                               core_serializers.RestrictedSerializerMixin,
                                               core_serializers.AugmentedSerializerMixin,
                                               serializers.HyperlinkedModelSerializer)):
    SERVICE_ACCOUNT_FIELDS = NotImplemented
    SERVICE_ACCOUNT_EXTRA_FIELDS = NotImplemented

    projects = BasicProjectSerializer(many=True, read_only=True)
    customer_native_name = serializers.ReadOnlyField(source='customer.native_name')
    settings = serializers.HyperlinkedRelatedField(
        queryset=models.ServiceSettings.objects.filter(shared=True),
        view_name='servicesettings-detail',
        lookup_field='uuid',
        allow_null=True)
    # if project is defined service will be automatically connected to projects customer
    # and SPL between service and project will be created
    project = serializers.HyperlinkedRelatedField(
        queryset=models.Project.objects.all().select_related('customer'),
        view_name='project-detail',
        lookup_field='uuid',
        allow_null=True,
        required=False,
        write_only=True)

    backend_url = serializers.URLField(max_length=200, allow_null=True, write_only=True, required=False)
    username = serializers.CharField(max_length=100, allow_null=True, write_only=True, required=False)
    password = serializers.CharField(max_length=100, allow_null=True, write_only=True, required=False)
    domain = serializers.CharField(max_length=200, allow_null=True, write_only=True, required=False)
    token = serializers.CharField(allow_null=True, write_only=True, required=False)
    certificate = serializers.FileField(allow_null=True, write_only=True, required=False)
    resources_count = serializers.SerializerMethodField()
    service_type = serializers.SerializerMethodField()
    state = serializers.SerializerMethodField()
    scope = core_serializers.GenericRelatedField(related_models=models.ResourceMixin.get_all_models(), required=False)
    tags = serializers.SerializerMethodField()
    quotas = quotas_serializers.BasicQuotaSerializer(many=True, read_only=True)

    shared = serializers.ReadOnlyField(source='settings.shared')
    error_message = serializers.ReadOnlyField(source='settings.error_message')
    terms_of_services = serializers.ReadOnlyField(source='settings.terms_of_services')
    homepage = serializers.ReadOnlyField(source='settings.homepage')
    geolocations = core_serializers.GeoLocationField(source='settings.geolocations', read_only=True)
    certifications = NestedServiceCertificationSerializer(many=True, read_only=True, source='settings.certifications')
    name = serializers.ReadOnlyField(source='settings.name')

    class Meta(object):
        model = NotImplemented
        fields = (
            'uuid', 'url', 'name', 'state', 'service_type', 'shared',
            'projects', 'project',
            'customer', 'customer_uuid', 'customer_name', 'customer_native_name', 'resources_count',
            'settings', 'settings_uuid', 'backend_url', 'username', 'password',
            'token', 'certificate', 'domain', 'terms_of_services', 'homepage',
            'certifications', 'geolocations', 'available_for_all', 'scope', 'tags', 'quotas',
        )
        settings_fields = ('backend_url', 'username', 'password', 'token', 'certificate', 'scope', 'domain')
        protected_fields = ('customer', 'settings', 'project') + settings_fields
        related_paths = ('customer', 'settings')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'customer': {'lookup_field': 'uuid'},
            'settings': {'lookup_field': 'uuid'},
        }

    def __new__(cls, *args, **kwargs):
        if cls.SERVICE_ACCOUNT_EXTRA_FIELDS is not NotImplemented:
            cls.Meta.fields += tuple(cls.SERVICE_ACCOUNT_EXTRA_FIELDS.keys())
            cls.Meta.protected_fields += tuple(cls.SERVICE_ACCOUNT_EXTRA_FIELDS.keys())
        return super(BaseServiceSerializer, cls).__new__(cls, *args, **kwargs)

    @staticmethod
    def eager_load(queryset, request=None):
        queryset = queryset.select_related('customer', 'settings')
        projects = models.Project.objects.all().only('uuid', 'name')
        return queryset.prefetch_related(django_models.Prefetch('projects', queryset=projects), 'quotas')

    def get_tags(self, service):
        return service.settings.get_tags()

    def get_filtered_field_names(self):
        return 'customer',

    def get_fields(self):
        fields = super(BaseServiceSerializer, self).get_fields()

        if self.Meta.model is not NotImplemented and 'settings' in fields:
            key = SupportedServices.get_model_key(self.Meta.model)
            fields['settings'].queryset = fields['settings'].queryset.filter(type=key)

        if self.SERVICE_ACCOUNT_FIELDS is not NotImplemented:
            # each service settings could be connected to scope
            self.SERVICE_ACCOUNT_FIELDS['scope'] = _('VM that contains service')
            for field in self.Meta.settings_fields:
                if field not in fields:
                    continue
                if field in self.SERVICE_ACCOUNT_FIELDS:
                    fields[field].help_text = self.SERVICE_ACCOUNT_FIELDS[field]
                else:
                    del fields[field]

        return fields

    def build_unknown_field(self, field_name, model_class):
        if self.SERVICE_ACCOUNT_EXTRA_FIELDS is not NotImplemented:
            if field_name in self.SERVICE_ACCOUNT_EXTRA_FIELDS:
                backend = SupportedServices.get_service_backend(self.Meta.model)
                kwargs = {
                    'write_only': True,
                    'required': False,
                    'allow_blank': True,
                    'help_text': self.SERVICE_ACCOUNT_EXTRA_FIELDS[field_name],
                }
                if hasattr(backend, 'DEFAULTS') and field_name in backend.DEFAULTS:
                    kwargs['help_text'] += ' (default: %s)' % json.dumps(backend.DEFAULTS[field_name])
                    kwargs['initial'] = backend.DEFAULTS[field_name]
                return serializers.CharField, kwargs

        return super(BaseServiceSerializer, self).build_unknown_field(field_name, model_class)

    def validate_empty_values(self, data):
        # required=False is ignored for settings FK, deal with it here
        if 'settings' not in data:
            data = data.copy()
            data['settings'] = None
        return super(BaseServiceSerializer, self).validate_empty_values(data)

    def validate(self, attrs):
        user = self.context['request'].user
        customer = attrs.get('customer') or self.instance.customer
        project = attrs.get('project')
        if project and project.customer != customer:
            raise serializers.ValidationError(
                _('Service cannot be connected to project that does not belong to services customer.'))

        settings = attrs.get('settings')
        if not user.is_staff:
            if not customer.has_user(user, models.CustomerRole.OWNER):
                raise exceptions.PermissionDenied()
            if not self.instance and settings and not settings.shared:
                if attrs.get('customer') != settings.customer:
                    raise serializers.ValidationError(_('Customer must match settings customer.'))

        if self.context['request'].method == 'POST':
            name = self.initial_data.get('name')
            if not name or not name.strip():
                raise serializers.ValidationError({'name': 'Name cannot be empty'})

            extra_fields = tuple()
            if self.SERVICE_ACCOUNT_EXTRA_FIELDS is not NotImplemented:
                extra_fields = tuple(self.SERVICE_ACCOUNT_EXTRA_FIELDS.keys())

            # Make shallow copy to protect from mutations
            settings_fields = self.Meta.settings_fields[:]
            create_settings = any([attrs.get(f) for f in settings_fields + extra_fields])
            if not settings and not create_settings:
                raise serializers.ValidationError(
                    _('Either service settings or credentials must be supplied.'))

            if create_settings:
                required = getattr(self.Meta, 'required_fields', tuple())
                for field in settings_fields:
                    if field in required and (field not in attrs or attrs[field] is None):
                        error = self.fields[field].error_messages['required']
                        raise serializers.ValidationError({field: six.text_type(error)})

                args = {f: attrs.get(f) for f in settings_fields if f in attrs}
                if extra_fields:
                    args['options'] = {f: attrs[f] for f in extra_fields if f in attrs}

                name = self.initial_data.get('name')
                if name is None:
                    raise serializers.ValidationError({'name': _('Name field is required.')})

                settings = models.ServiceSettings(
                    type=SupportedServices.get_model_key(self.Meta.model),
                    name=name,
                    customer=customer,
                    **args)

                try:
                    backend = settings.get_backend()
                    backend.ping(raise_exception=True)
                except ServiceBackendError as e:
                    raise serializers.ValidationError(_('Wrong settings: %s.') % e)
                except ServiceBackendNotImplemented:
                    pass

                self._validate_settings(settings)

                settings.save()
                executors.ServiceSettingsCreateExecutor.execute(settings)
                attrs['settings'] = settings

            for f in settings_fields + extra_fields:
                if f in attrs:
                    del attrs[f]

        return attrs

    def _validate_settings(self, settings):
        pass

    def get_resources_count(self, service):
        return self.get_resources_count_map[service.pk]

    @cached_property
    def get_resources_count_map(self):
        resource_models = SupportedServices.get_service_resources(self.Meta.model)
        resource_models = set(resource_models) - set(models.SubResource.get_all_models())
        counts = defaultdict(lambda: 0)
        user = self.context['request'].user
        for model in resource_models:
            service_path = model.Permissions.service_path
            if isinstance(self.instance, list):
                query = {service_path + '__in': self.instance}
            else:
                query = {service_path: self.instance}
            queryset = filter_queryset_for_user(model.objects.all(), user)
            rows = queryset.filter(**query).values(service_path) \
                .annotate(count=django_models.Count('id'))
            for row in rows:
                service_id = row[service_path]
                counts[service_id] += row['count']
        return counts

    def get_service_type(self, obj):
        return SupportedServices.get_name_for_model(obj)

    def get_state(self, obj):
        return obj.settings.get_state_display()

    def create(self, attrs):
        project = attrs.pop('project', None)
        service = super(BaseServiceSerializer, self).create(attrs)
        spl_model = service.projects.through
        if project and not spl_model.objects.filter(project=project, service=service).exists():
            spl_model.objects.create(project=project, service=service)
        return service


class BaseServiceProjectLinkSerializer(PermissionFieldFilteringMixin,
                                       core_serializers.AugmentedSerializerMixin,
                                       serializers.HyperlinkedModelSerializer):
    project = serializers.HyperlinkedRelatedField(
        queryset=models.Project.objects.all(),
        view_name='project-detail',
        lookup_field='uuid')

    service_name = serializers.ReadOnlyField(source='service.settings.name')
    quotas = quotas_serializers.BasicQuotaSerializer(many=True, read_only=True)

    class Meta(object):
        model = NotImplemented
        fields = (
            'url',
            'project', 'project_name', 'project_uuid',
            'service', 'service_uuid', 'service_name', 'quotas',
        )
        related_paths = ('project', 'service')
        extra_kwargs = {
            'service': {'lookup_field': 'uuid', 'view_name': NotImplemented},
        }

    def get_filtered_field_names(self):
        return 'project', 'service'

    def validate(self, attrs):
        if attrs['service'].customer != attrs['project'].customer:
            raise serializers.ValidationError(_("Service customer doesn't match project customer."))

        # XXX: Consider adding unique key (service, project) to the model instead
        if self.Meta.model.objects.filter(service=attrs['service'], project=attrs['project']).exists():
            raise serializers.ValidationError(_('This service project link already exists.'))

        return attrs


class ResourceSerializerMetaclass(serializers.SerializerMetaclass):
    """ Build a list of supported resource via serializers definition.
        See SupportedServices for details.
    """

    def __new__(cls, name, bases, args):
        serializer = super(ResourceSerializerMetaclass, cls).__new__(cls, name, bases, args)
        if 'Meta' in args:
            SupportedServices.register_resource_serializer(args['Meta'].model, serializer)
        return serializer


class BasicResourceSerializer(serializers.Serializer):
    uuid = serializers.ReadOnlyField()
    name = serializers.ReadOnlyField()
    resource_type = serializers.SerializerMethodField()

    def get_resource_type(self, resource):
        return SupportedServices.get_name_for_model(resource)


class ManagedResourceSerializer(BasicResourceSerializer):
    project_name = serializers.ReadOnlyField(source='service_project_link.project.name')
    project_uuid = serializers.ReadOnlyField(source='service_project_link.project.uuid')

    customer_uuid = serializers.ReadOnlyField(source='service_project_link.project.customer.uuid')
    customer_name = serializers.ReadOnlyField(source='service_project_link.project.customer.name')


class TagList(list):
    """
    This class serializes tags as JSON list as the last step of serialization process.
    """

    def __str__(self):
        return json.dumps(self)


class TagSerializer(serializers.Serializer):
    """
    This serializer updates tags field using django-taggit API.
    """

    def create(self, validated_data):
        if 'tags' in validated_data:
            tags = validated_data.pop('tags')
            instance = super(TagSerializer, self).create(validated_data)
            instance.tags.set(*tags)
        else:
            instance = super(TagSerializer, self).create(validated_data)
        return instance

    def update(self, instance, validated_data):
        if 'tags' in validated_data:
            tags = validated_data.pop('tags')
            instance = super(TagSerializer, self).update(instance, validated_data)
            instance.tags.set(*tags)
        else:
            instance = super(TagSerializer, self).update(instance, validated_data)
        return instance


class TagListSerializerField(serializers.Field):
    child = serializers.CharField()
    default_error_messages = {
        'not_a_list': _('Expected a list of items but got type "{input_type}".'),
        'invalid_json': _('Invalid json list. A tag list submitted in string form must be valid json.'),
        'not_a_str': _('All list items must be of string type.')
    }

    def to_internal_value(self, value):
        if isinstance(value, six.string_types):
            if not value:
                value = '[]'
            try:
                value = json.loads(value)
            except ValueError:
                self.fail('invalid_json')

        if not isinstance(value, list):
            self.fail('not_a_list', input_type=type(value).__name__)

        for s in value:
            if not isinstance(s, six.string_types):
                self.fail('not_a_str')

            self.child.run_validation(s)

        return value

    def get_attribute(self, instance):
        """
        Fetch tags from cache defined in TagMixin.
        """
        return instance.get_tags()

    def to_representation(self, value):
        if not isinstance(value, TagList):
            value = TagList(value)
        return value


class BaseResourceSerializer(six.with_metaclass(ResourceSerializerMetaclass,
                                                core_serializers.RestrictedSerializerMixin,
                                                MonitoringSerializerMixin,
                                                PermissionFieldFilteringMixin,
                                                core_serializers.AugmentedSerializerMixin,
                                                TagSerializer,
                                                serializers.HyperlinkedModelSerializer)):
    state = serializers.ReadOnlyField(source='get_state_display')

    project = serializers.HyperlinkedRelatedField(
        queryset=models.Project.objects.all(),
        view_name='project-detail',
        lookup_field='uuid',
        allow_null=True,
        required=False,
    )

    project_name = serializers.ReadOnlyField(source='service_project_link.project.name')
    project_uuid = serializers.ReadOnlyField(source='service_project_link.project.uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name=NotImplemented,
        queryset=NotImplemented,
        allow_null=True,
        required=False,
    )

    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name=NotImplemented,
        read_only=True,
        lookup_field='uuid')

    service_name = serializers.ReadOnlyField(source='service_project_link.service.settings.name')
    service_uuid = serializers.ReadOnlyField(source='service_project_link.service.uuid')

    service_settings = serializers.HyperlinkedRelatedField(
        queryset=models.ServiceSettings.objects.all(),
        view_name='servicesettings-detail',
        lookup_field='uuid',
        allow_null=True,
        required=False,
    )
    service_settings_uuid = serializers.ReadOnlyField(source='service_project_link.service.settings.uuid')
    service_settings_state = serializers.ReadOnlyField(
        source='service_project_link.service.settings.human_readable_state')
    service_settings_error_message = serializers.ReadOnlyField(
        source='service_project_link.service.settings.error_message')

    customer = serializers.HyperlinkedRelatedField(
        source='service_project_link.project.customer',
        view_name='customer-detail',
        read_only=True,
        lookup_field='uuid')

    customer_name = serializers.ReadOnlyField(source='service_project_link.project.customer.name')
    customer_abbreviation = serializers.ReadOnlyField(source='service_project_link.project.customer.abbreviation')
    customer_native_name = serializers.ReadOnlyField(source='service_project_link.project.customer.native_name')

    created = serializers.DateTimeField(read_only=True)
    resource_type = serializers.SerializerMethodField()

    tags = TagListSerializerField(required=False)
    access_url = serializers.SerializerMethodField()
    is_link_valid = serializers.BooleanField(
        source='service_project_link.is_valid',
        read_only=True,
        help_text=_('True if resource is originated from a service that satisfies an associated project requirements.'))

    class Meta(object):
        model = NotImplemented
        fields = MonitoringSerializerMixin.Meta.fields + (
            'url', 'uuid', 'name', 'description',
            'service', 'service_name', 'service_uuid',
            'service_settings', 'service_settings_uuid',
            'service_settings_state', 'service_settings_error_message',
            'project', 'project_name', 'project_uuid',
            'customer', 'customer_name', 'customer_native_name', 'customer_abbreviation',
            'tags', 'error_message',
            'resource_type', 'state', 'created', 'service_project_link', 'backend_id',
            'access_url', 'is_link_valid',
        )
        protected_fields = ('service', 'service_project_link', 'project', 'service_settings')
        read_only_fields = ('error_message', 'backend_id')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    def get_filtered_field_names(self):
        return 'service_project_link',

    def get_resource_type(self, obj):
        return SupportedServices.get_name_for_model(obj)

    def get_resource_fields(self):
        return [f.name for f in self.Meta.model._meta.get_fields()]

    # an optional generic URL for accessing a resource
    def get_access_url(self, obj):
        return obj.get_access_url()

    @staticmethod
    def eager_load(queryset, request=None):
        return (
            queryset
            .select_related(
                'service_project_link',
                'service_project_link__service',
                'service_project_link__service__settings',
                'service_project_link__project',
                'service_project_link__project__customer',
            ).prefetch_related('service_project_link__service__settings__certifications',
                               'service_project_link__project__certifications')
        )

    def get_fields(self):
        fields = super(BaseResourceSerializer, self).get_fields()
        # skip validation on object update
        if not self.instance:
            service_type = SupportedServices.get_model_key(self.Meta.model)
            if 'service_settings' in fields and not fields['service_settings'].read_only:
                queryset = fields['service_settings'].queryset.filter(type=service_type)
                fields['service_settings'].queryset = queryset
        return fields

    def validate(self, attrs):
        # skip validation on object update
        if self.instance:
            return attrs

        service_settings = attrs.pop('service_settings', None)
        project = attrs.pop('project', None)
        service_project_link = attrs.get('service_project_link')

        if not service_project_link:
            if service_settings and project:
                spl_model = self.Meta.model.service_project_link.field.remote_field.model
                try:
                    service_project_link = spl_model.objects.get(
                        service__settings=service_settings,
                        project=project,
                    )
                    attrs['service_project_link'] = service_project_link
                except django_exceptions.ObjectDoesNotExist:
                    raise serializers.ValidationError(
                        _('You are not allowed to provision resource in current project using this provider. '
                          'Please specify another value for project and service_settings fields.')
                    )
            else:
                raise serializers.ValidationError(
                    _('Either service_project_link or service_settings and project should be specified.')
                )

        if not service_project_link.is_valid:
            raise serializers.ValidationError({
                'service_project_link': service_project_link.validation_message
            })

        return attrs

    @transaction.atomic
    def create(self, validated_data):
        data = validated_data.copy()
        fields = self.get_resource_fields()
        # Remove `virtual` properties which ain't actually belong to the model
        for prop in data.keys():
            if prop not in fields:
                del data[prop]

        resource = super(BaseResourceSerializer, self).create(data)
        resource.increase_backend_quotas_usage()
        return resource


class BaseResourceActionSerializer(BaseResourceSerializer):
    project = serializers.HyperlinkedRelatedField(
        view_name='project-detail',
        lookup_field='uuid',
        read_only=True,
    )
    service_settings = serializers.HyperlinkedRelatedField(
        view_name='servicesettings-detail',
        lookup_field='uuid',
        read_only=True,
    )

    class Meta(BaseResourceSerializer.Meta):
        pass


class SummaryResourceSerializer(core_serializers.BaseSummarySerializer):
    @classmethod
    def get_serializer(cls, model):
        return SupportedServices.get_resource_serializer(model)


class SummaryServiceSerializer(core_serializers.BaseSummarySerializer):
    @classmethod
    def get_serializer(cls, model):
        return SupportedServices.get_service_serializer(model)


class BaseResourceImportSerializer(PermissionFieldFilteringMixin,
                                   core_serializers.AugmentedSerializerMixin,
                                   serializers.HyperlinkedModelSerializer):
    backend_id = serializers.CharField(write_only=True)
    project = serializers.HyperlinkedRelatedField(
        queryset=models.Project.objects.all(),
        view_name='project-detail',
        lookup_field='uuid',
        write_only=True)

    state = serializers.ReadOnlyField(source='get_state_display')
    created = serializers.DateTimeField(read_only=True)
    import_history = serializers.BooleanField(
        default=True, write_only=True, help_text=_('Import historical resource usage.'))

    class Meta(object):
        model = NotImplemented
        fields = (
            'url', 'uuid', 'name', 'state', 'created',
            'backend_id', 'project', 'import_history'
        )
        read_only_fields = ('name',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    def get_filtered_field_names(self):
        return 'project',

    def get_fields(self):
        fields = super(BaseResourceImportSerializer, self).get_fields()
        # Context doesn't have service during schema generation
        if 'service' in self.context:
            fields['project'].queryset = self.context['service'].projects.all()

        return fields

    def validate(self, attrs):
        if self.Meta.model.objects.filter(backend_id=attrs['backend_id']).exists():
            raise serializers.ValidationError(
                {'backend_id': _('This resource is already linked to Waldur.')})

        spl_class = SupportedServices.get_related_models(self.Meta.model)['service_project_link']
        spl = spl_class.objects.get(service=self.context['service'], project=attrs['project'])
        attrs['service_project_link'] = spl

        return attrs

    def create(self, validated_data):
        validated_data.pop('project')
        return super(BaseResourceImportSerializer, self).create(validated_data)


class VirtualMachineSerializer(BaseResourceSerializer):
    external_ips = serializers.ListField(
        child=serializers.IPAddressField(protocol='ipv4'),
        read_only=True,
    )
    internal_ips = serializers.ListField(
        child=serializers.IPAddressField(protocol='ipv4'),
        read_only=True,
    )

    ssh_public_key = serializers.HyperlinkedRelatedField(
        view_name='sshpublickey-detail',
        lookup_field='uuid',
        queryset=core_models.SshPublicKey.objects.all(),
        required=False,
        write_only=True)

    class Meta(BaseResourceSerializer.Meta):
        fields = BaseResourceSerializer.Meta.fields + (
            'start_time', 'cores', 'ram', 'disk', 'min_ram', 'min_disk',
            'ssh_public_key', 'user_data', 'external_ips', 'internal_ips',
            'latitude', 'longitude', 'key_name', 'key_fingerprint', 'image_name'
        )
        read_only_fields = BaseResourceSerializer.Meta.read_only_fields + (
            'start_time', 'cores', 'ram', 'disk', 'min_ram', 'min_disk',
            'external_ips', 'internal_ips',
            'latitude', 'longitude', 'key_name', 'key_fingerprint', 'image_name'
        )
        protected_fields = BaseResourceSerializer.Meta.protected_fields + (
            'user_data', 'ssh_public_key'
        )

    def get_fields(self):
        fields = super(VirtualMachineSerializer, self).get_fields()
        if 'request' in self.context:
            user = self.context['request'].user
            ssh_public_key = fields.get('ssh_public_key')
            if ssh_public_key:
                ssh_public_key.query_params = {'user_uuid': user.uuid.hex}
                if not user.is_staff:
                    visible_users = list(filter_visible_users(User.objects.all(), user))
                    subquery = Q(user__in=visible_users) | Q(is_shared=True)
                    ssh_public_key.queryset = ssh_public_key.queryset.filter(subquery)
        return fields

    def create(self, validated_data):
        if 'image' in validated_data:
            validated_data['image_name'] = validated_data['image'].name
        return super(VirtualMachineSerializer, self).create(validated_data)


class PropertySerializerMetaclass(serializers.SerializerMetaclass):
    """ Build a list of supported properties via serializers definition.
        See SupportedServices for details.
    """

    def __new__(cls, name, bases, args):
        SupportedServices.register_property(args['Meta'].model)
        return super(PropertySerializerMetaclass, cls).__new__(cls, name, bases, args)


class BasePropertySerializer(six.with_metaclass(PropertySerializerMetaclass,
                                                core_serializers.AugmentedSerializerMixin,
                                                serializers.HyperlinkedModelSerializer)):
    class Meta(object):
        model = NotImplemented


class AggregateSerializer(serializers.Serializer):
    MODEL_NAME_CHOICES = (
        ('project', 'project'),
        ('customer', 'customer'),
    )
    MODEL_CLASSES = {
        'project': models.Project,
        'customer': models.Customer,
    }

    aggregate = serializers.ChoiceField(choices=MODEL_NAME_CHOICES, default='customer')
    uuid = serializers.CharField(allow_null=True, default=None)

    def get_aggregates(self, user):
        model = self.MODEL_CLASSES[self.data['aggregate']]
        queryset = filter_queryset_for_user(model.objects.all(), user)

        if 'uuid' in self.data and self.data['uuid']:
            queryset = queryset.filter(uuid=self.data['uuid'])
        return queryset

    def get_projects(self, user):
        queryset = self.get_aggregates(user)

        if self.data['aggregate'] == 'project':
            return queryset.all()
        else:
            queryset = models.Project.objects.filter(customer__in=list(queryset))
            return filter_queryset_for_user(queryset, user)

    def get_service_project_links(self, user):
        projects = self.get_projects(user)
        return [model.objects.filter(project__in=projects)
                for model in models.ServiceProjectLink.get_all_models()]


class PrivateCloudSerializer(BaseResourceSerializer):
    extra_configuration = serializers.JSONField(read_only=True)

    class Meta(BaseResourceSerializer.Meta):
        fields = BaseResourceSerializer.Meta.fields + ('extra_configuration',)


class DivisionSerializer(serializers.HyperlinkedModelSerializer):
    type = serializers.ReadOnlyField(source='type.name')
    parent_uuid = serializers.ReadOnlyField(source='parent.uuid')

    class Meta(object):
        model = models.Division
        fields = ('uuid', 'url', 'name', 'type', 'parent_uuid', 'parent')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'parent': {'lookup_field': 'uuid'},
        }



import copy
import functools
import heapq
import itertools

from django.contrib.contenttypes.models import ContentType
from django.db import models
import six


class GenericKeyMixin(object):
    """
    Filtering by generic key field

    Support filtering by:
     - generic key directly: <generic_key_name>=<value>
     - is generic key null: <generic_key_name>__isnull=True|False
    """

    def __init__(
            self, generic_key_field='scope',
            object_id_field='object_id', content_type_field='content_type', available_models=(), **kwargs):
        super(GenericKeyMixin, self).__init__(**kwargs)
        self.generic_key_field = generic_key_field
        self.object_id_field = object_id_field
        self.content_type_field = content_type_field
        self.available_models = available_models

    def _preprocess_kwargs(self, initial_kwargs):
        """ Replace generic key related attribute with filters by object_id and content_type fields """
        kwargs = initial_kwargs.copy()
        generic_key_related_kwargs = self._get_generic_key_related_kwargs(initial_kwargs)
        for key, value in generic_key_related_kwargs.items():
            # delete old kwarg that was related to generic key
            del kwargs[key]
            try:
                suffix = key.split('__')[1]
            except IndexError:
                suffix = None
            # add new kwargs that related to object_id and content_type fields
            new_kwargs = self._get_filter_object_id_and_content_type_filter_kwargs(value, suffix)
            kwargs.update(new_kwargs)

        return kwargs

    def _get_generic_key_related_kwargs(self, initial_kwargs):
        # Skip fields like scope_customer
        return {key: value for key, value in initial_kwargs.items()
                if key.startswith(self.generic_key_field + '__') or key == self.generic_key_field}

    def _get_filter_object_id_and_content_type_filter_kwargs(self, generic_key_value, suffix=None):
        kwargs = {}
        if suffix is None:
            kwargs[self.object_id_field] = generic_key_value.id
            generic_key_content_type = ContentType.objects.get_for_model(generic_key_value)
            kwargs[self.content_type_field] = generic_key_content_type
        elif suffix == 'in':
            kwargs[self.object_id_field + '__in'] = [obj.id for obj in generic_key_value]
            content_type = ContentType.objects.get_for_model(generic_key_value[0]) if generic_key_value else None
            kwargs[self.content_type_field] = content_type
        elif suffix == 'isnull':
            kwargs[self.object_id_field + '__isnull'] = generic_key_value
            kwargs[self.content_type_field + '__isnull'] = generic_key_value
        return kwargs

    def filter(self, *args, **kwargs):
        kwargs = self._preprocess_kwargs(kwargs)
        return super(GenericKeyMixin, self).filter(*args, **kwargs)

    def get(self, *args, **kwargs):
        kwargs = self._preprocess_kwargs(kwargs)
        return super(GenericKeyMixin, self).get(*args, **kwargs)

    def get_or_create(self, *args, **kwargs):
        kwargs = self._preprocess_kwargs(kwargs)
        return super(GenericKeyMixin, self).get_or_create(*args, **kwargs)


class SummaryQuerySet(object):
    """ Fake queryset that emulates union of different models querysets """

    def __init__(self, summary_models):
        self.querysets = [model.objects.all() for model in summary_models]
        self._order_by = None

    def filter(self, *args, **kwargs):
        self.querysets = [qs.filter(*copy.deepcopy(args), **copy.deepcopy(kwargs)) for qs in self.querysets]
        return self

    def exclude(self, *args, **kwargs):
        self.querysets = [qs.exclude(*copy.deepcopy(args), **copy.deepcopy(kwargs)) for qs in self.querysets]
        return self

    def distinct(self, *args, **kwargs):
        self.querysets = [qs.distinct(*copy.deepcopy(args), **copy.deepcopy(kwargs)) for qs in self.querysets]
        return self

    def order_by(self, order_by):
        self._order_by = order_by
        self.querysets = [qs.order_by(copy.deepcopy(order_by)) for qs in self.querysets]
        return self

    def count(self):
        return sum([qs.count() for qs in self.querysets])

    def all(self):
        return self

    def none(self):
        try:
            return self.querysets[0].none()
        except IndexError:
            return

    def __getitem__(self, val):
        chained_querysets = self._get_chained_querysets()
        if isinstance(val, slice):
            return list(itertools.islice(chained_querysets, val.start, val.stop))
        else:
            try:
                return next(itertools.islice(chained_querysets, val, val + 1))
            except StopIteration:
                raise IndexError

    def __len__(self):
        return sum([q.count() for q in self.querysets])

    def _get_chained_querysets(self):
        if self._order_by:
            return self._merge([qs.iterator() for qs in self.querysets], compared_attr=self._order_by)
        else:
            return itertools.chain(*[qs.iterator() for qs in self.querysets])

    def _merge(self, subsequences, compared_attr='pk'):

        @functools.total_ordering
        class Compared(object):
            """ Order objects by their attributes, reverse ordering if <reverse> is True """

            def __init__(self, obj, attr, reverse=False):
                self.attr = functools.reduce(Compared.get_obj_attr, attr.split("__"), obj)
                if isinstance(self.attr, six.string_types):
                    self.attr = self.attr.lower()
                self.reverse = reverse

            @staticmethod
            def get_obj_attr(obj, attr):
                # for m2m relationship support - get first instance of manager.
                # for example: get first project group if resource has to be ordered by project groups.
                if isinstance(obj, models.Manager):
                    obj = obj.first()
                return getattr(obj, attr) if obj else None

            def __eq__(self, other):
                return self.attr == other.attr

            def __le__(self, other):
                # In MySQL NULL values come *first* with ascending sort order.
                # We use the same behaviour.
                if self.attr is None:
                    return not self.reverse
                elif other.attr is None:
                    return self.reverse
                else:
                    return self.attr < other.attr if not self.reverse else self.attr >= other.attr

        reverse = compared_attr.startswith('-')
        if reverse:
            compared_attr = compared_attr[1:]

        # prepare a heap whose items are
        # (compared, current-value, iterator), one each per (non-empty) subsequence
        # <compared> is used for model instances comparison based on given attribute
        heap = []
        for subseq in subsequences:
            iterator = iter(subseq)
            for current_value in iterator:
                # subseq is not empty, therefore add this subseq's item to the list
                heapq.heappush(
                    heap, (Compared(current_value, compared_attr, reverse=reverse), current_value, iterator))
                break

        while heap:
            # get and yield lowest current value (and corresponding iterator)
            _, current_value, iterator = heap[0]
            yield current_value
            for current_value in iterator:
                # subseq is not finished, therefore add this subseq's item back into the priority queue
                heapq.heapreplace(
                    heap, (Compared(current_value, compared_attr, reverse=reverse), current_value, iterator))
                break
            else:
                # subseq has been exhausted, therefore remove it from the queue
                heapq.heappop(heap)



import django_filters

from waldur_core.structure import filters as structure_filters

from . import models


class ImageFilter(structure_filters.BaseServicePropertyFilter):

    class Meta:
        model = models.Image
        fields = structure_filters.BaseServicePropertyFilter.Meta.fields + ('region',)

    region = django_filters.UUIDFilter(name='region__uuid')


class SizeFilter(structure_filters.BaseServicePropertyFilter):

    class Meta:
        model = models.Size
        fields = structure_filters.BaseServicePropertyFilter.Meta.fields + ('region',)

    region = django_filters.UUIDFilter(name='regions__uuid')


class RegionFilter(structure_filters.BaseServicePropertyFilter):

    class Meta(structure_filters.BaseServicePropertyFilter.Meta):
        model = models.Region


class InstanceFilter(structure_filters.BaseResourceFilter):
    external_ip = django_filters.CharFilter(name='public_ips')

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Instance



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-11-02 20:42
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.core.fields
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0031_supportcustomer'),
    ]

    operations = [
        migrations.CreateModel(
            name='RequestType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.IntegerField(unique=True)),
                ('issue_type_name', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import django_filters

from waldur_core.core import filters as core_filters
from waldur_core.structure import filters as structure_filters

from . import models


class SlurmServiceProjectLinkFilter(structure_filters.BaseServiceProjectLinkFilter):
    service = core_filters.URLFilter(view_name='slurm-detail', name='service__uuid')

    class Meta(structure_filters.BaseServiceProjectLinkFilter.Meta):
        model = models.SlurmServiceProjectLink


class AllocationFilter(structure_filters.BaseResourceFilter):
    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Allocation
        fields = structure_filters.BaseResourceFilter.Meta.fields + ('is_active',)


class AllocationUsageFilter(django_filters.FilterSet):
    user = core_filters.URLFilter(view_name='user-detail', name='user__uuid')
    user_uuid = django_filters.UUIDFilter(name='user__uuid')

    allocation = core_filters.URLFilter(view_name='slurm-allocation-detail', name='allocation__uuid')
    allocation_uuid = django_filters.UUIDFilter(name='allocation__uuid')

    class Meta(object):
        model = models.AllocationUsage
        fields = ('year', 'month')



import json
import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand, CommandError
from rest_framework import status
from rest_framework.test import APIClient
import six

User = get_user_model()


class Command(BaseCommand):
    help = "Exports Waldur API Documentation in JSON format"

    def add_arguments(self, parser):
        parser.add_argument(
            '-o', '--output',
            dest='output', default=None,
            help='Specifies file to which the output is written. The output will be printed to stdout by default.',
        )

    def handle(self, *args, **options):
        # Rise logging level to prevent redundant log messages
        logging.disable(logging.CRITICAL)

        if not settings.DEBUG and not settings.ALLOWED_HOSTS:
            raise CommandError('ALLOWED_HOSTS should not be empty in settings.py file.')

        if settings.ALLOWED_HOSTS and settings.ALLOWED_HOSTS[0] != '*':
            host = settings.ALLOWED_HOSTS[0]
        else:
            host = '127.0.0.1'

        client = APIClient(HTTP_HOST=host)
        user, _ = User.objects.get_or_create(username='waldur_docs_exporter', is_staff=True)
        client.force_authenticate(user=user)
        response = client.get('/docs/?format=openapi')
        user.delete()

        if response.status_code != status.HTTP_200_OK:
            raise CommandError('Failed to get response from the server')

        data = json.loads(response.content)
        if options['output'] is None:
            self.stdout.write(six.text_type(data))
        else:
            with open(options['output'], 'w') as output_file:
                json.dump(data, output_file)

        # return logging level back
        logging.disable(logging.NOTSET)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-10-24 13:23
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0041_project_resource_count'),
    ]

    operations = [
        migrations.AddField(
            model_name='order',
            name='_file',
            field=models.TextField(blank=True, editable=False),
        ),
    ]



import datetime
from decimal import Decimal
import logging

from celery import shared_task
from dateutil.relativedelta import relativedelta
from django.contrib.contenttypes.models import ContentType
from django.core.mail import send_mail

from waldur_core.core import tasks as core_tasks, utils as core_utils
from waldur_core.monitoring.models import ResourceItem, ResourceSla, ResourceSlaStateTransition
from waldur_core.monitoring.utils import format_period

from .backend import ZabbixBackendError
from .models import Host, ITService, Item, SlaHistory

logger = logging.getLogger(__name__)


@shared_task(name='waldur_core.zabbix.pull_sla')
def pull_sla(host_uuid):
    """
    Pull SLAs for given Zabbix host for all time of its existence in Zabbix
    """
    try:
        host = Host.objects.get(uuid=host_uuid)
    except Host.DoesNotExist:
        logger.warning('Unable to pull SLA for host with UUID %s, because it is gone', host_uuid)
        return

    try:
        itservice = ITService.objects.get(host=host, is_main=True)
    except ITService.DoesNotExist:
        logger.warning('Unable to pull SLA for host with UUID %s, because IT service does not exist', host_uuid)
        return

    backend = itservice.get_backend()

    try:
        # Get dates of first and last service alarm
        min_dt, max_dt = backend.get_sla_range(itservice.backend_id)
    except ZabbixBackendError as e:
        logger.warning('Unable to pull SLA for host with with UUID %s because of database error: %s', host_uuid, e)
        return

    # Shift date to beginning of the month
    current_point = min_dt.replace(day=1)
    while current_point <= max_dt:
        period = format_period(current_point)
        start_time = core_utils.datetime_to_timestamp(current_point)
        current_point += relativedelta(months=+1)
        end_time = core_utils.datetime_to_timestamp(min(max_dt, current_point))
        update_itservice_sla.delay(itservice.pk, period, start_time, end_time)

    logger.debug('Successfully pulled SLA for host with with UUID %s', host_uuid)


@shared_task(name='waldur_core.zabbix.update_sla')
def update_sla(sla_type):
    if sla_type not in ('yearly', 'monthly'):
        logger.error('Requested unknown SLA type: %s' % sla_type)
        return

    dt = datetime.datetime.now()

    if sla_type == 'yearly':
        period = dt.year
        start_time = int(datetime.datetime.strptime('01/01/%s' % dt.year, '%d/%m/%Y').strftime("%s"))
    else:  # it's a monthly SLA update
        period = format_period(dt)
        month_start = datetime.datetime.strptime('01/%s/%s' % (dt.month, dt.year), '%d/%m/%Y')
        start_time = int(month_start.strftime("%s"))

    end_time = int(dt.strftime("%s"))

    for itservice in ITService.objects.all():
        update_itservice_sla.delay(itservice.pk, period, start_time, end_time)


@shared_task
def update_itservice_sla(itservice_pk, period, start_time, end_time):
    logger.debug('Updating SLAs for IT Service with PK %s. Period: %s, start_time: %s, end_time: %s',
                 itservice_pk, period, start_time, end_time)

    try:
        itservice = ITService.objects.get(pk=itservice_pk)
    except ITService.DoesNotExist:
        logger.warning('Unable to update SLA for IT Service with PK %s, because it is gone', itservice_pk)
        return

    backend = itservice.host.get_backend()

    try:
        current_sla = backend.get_sla(itservice.backend_id, start_time, end_time)
        entry, _ = SlaHistory.objects.get_or_create(itservice=itservice, period=period)
        entry.value = Decimal(current_sla)
        entry.save()

        # Save SLA if IT service is marked as main for host
        if itservice.host and itservice.host.scope and itservice.is_main:
            ResourceSla.objects.update_or_create(
                object_id=itservice.host.scope.id,
                content_type=ContentType.objects.get_for_model(itservice.host.scope),
                period=period,
                defaults={
                    'value': current_sla,
                    'agreed_value': itservice.agreed_sla
                }
            )

        if itservice.backend_trigger_id:
            # update connected events
            events = backend.get_trigger_events(itservice.backend_trigger_id, start_time, end_time)
            for event in events:
                event_state = 'U' if int(event['value']) == 0 else 'D'
                entry.events.get_or_create(
                    timestamp=int(event['timestamp']),
                    state=event_state
                )

                if itservice.host and itservice.host.scope and itservice.is_main:
                    ResourceSlaStateTransition.objects.get_or_create(
                        scope=itservice.host.scope,
                        period=period,
                        timestamp=int(event['timestamp']),
                        state=int(event['value']) == 0
                    )
    except ZabbixBackendError as e:
        logger.warning(
            'Unable to update SLA for IT Service %s (ID: %s). Reason: %s', itservice.name, itservice.backend_id, e)
    logger.debug('Successfully updated SLA for IT Service %s (ID: %s)', itservice.name, itservice.backend_id)


@shared_task(name='waldur_core.zabbix.update_monitoring_items')
def update_monitoring_items():
    """
    Regularly update value of monitored resources
    """
    hosts = Host.objects.filter(object_id__isnull=False, state=Host.States.OK)
    for host in hosts:
        for config in host.MONITORING_ITEMS_CONFIGS:
            update_host_scope_monitoring_items.delay(host.uuid.hex,
                                                     zabbix_item_key=config['zabbix_item_key'],
                                                     monitoring_item_name=config['monitoring_item_name'])
    logger.debug('Successfully scheduled monitoring data update for zabbix hosts.')


@shared_task
def update_host_scope_monitoring_items(host_uuid, zabbix_item_key, monitoring_item_name):
    host = Host.objects.get(uuid=host_uuid)
    if host.scope is None:
        return None
    value = None
    if Item.objects.filter(template__hosts=host, key=zabbix_item_key).exists():
        value = host.get_backend().get_item_last_value(host.backend_id, key=zabbix_item_key)
        ResourceItem.objects.update_or_create(
            object_id=host.scope.id,
            content_type=ContentType.objects.get_for_model(host.scope),
            name=monitoring_item_name,
            defaults={'value': value}
        )
        logger.debug('Successfully updated monitoring item %s for host %s (%s). Current value: %s.',
                     monitoring_item_name, host.visible_name, host.uuid.hex, value)
    else:
        logger.debug('Host %s (UUID: %s) does not have monitoring item %s.',
                     host.visible_name, host.uuid.hex, monitoring_item_name)
    return value


@shared_task(max_retries=60, default_retry_delay=60)
def after_creation_monitoring_item_update(host_uuid, config):
    item_value = update_host_scope_monitoring_items(
        host_uuid, config['zabbix_item_key'], config['monitoring_item_name'])
    return item_value in config.get('after_creation_update_terminate_values', []) or item_value is None


class SMSTask(core_tasks.Task):
    """ Send SMS to given mobile number based on service settings or django settings """

    def execute(self, settings, message, phone):
        sender = settings.get_option('sms_email_from')
        recipient = settings.get_option('sms_email_rcpt')

        if sender and recipient and '{phone}' in recipient:
            send_mail('', message, sender, [recipient.format(phone=phone)], fail_silently=True)
        else:
            logger.warning('SMS was not sent, because `sms_email_from` and `sms_email_rcpt` '
                           'were not configured properly.')


@shared_task(name='waldur_core.zabbix.pull_hosts')
def pull_hosts():
    pullable_hosts = Host.objects.exclude(backend_id='')  # Cannot pull hosts without backend_id
    for host in pullable_hosts.filter(state=Host.States.ERRED):
        serialized_host = core_utils.serialize_instance(host)
        core_tasks.BackendMethodTask().apply_async(
            args=(serialized_host, 'pull_host'),
            link=core_tasks.RecoverTask().si(serialized_host),
            link_error=core_tasks.ErrorMessageTask().s(serialized_host),
        )
    for host in Host.objects.filter(state=Host.States.OK):
        serialized_host = core_utils.serialize_instance(host)
        core_tasks.BackendMethodTask().apply_async(
            args=(serialized_host, 'pull_host'),
            link_error=core_tasks.ErrorStateTransitionTask().s(serialized_host)
        )


class UpdateSettingsCredentials(core_tasks.Task):

    def execute(self, service_settings, serialized_user):
        user = core_utils.deserialize_instance(serialized_user)
        service_settings.password = user.password
        service_settings.save()



from __future__ import unicode_literals

from django.utils.module_loading import import_string

from waldur_core.core import WaldurExtension


# This function is needed in order to break circular dependency
def login_failed(request, args, **kwargs):
    return import_string('waldur_auth_openid.views.login_failed')(request, args, **kwargs)


class WaldurAuthOpenIDExtension(WaldurExtension):
    class Settings:

        # the library python-openid does not support a json session serializer
        # <openid.yadis.manager.YadisServiceManager> is not JSON serializable
        # https://github.com/openid/python-openid/issues/17
        SESSION_SERIALIZER = 'django.contrib.sessions.serializers.PickleSerializer'

        # Should users be created when new OpenIDs are used to log in?
        OPENID_CREATE_USERS = True

        # When logging in again, should we overwrite user details based on
        # data received via Simple Registration?
        OPENID_UPDATE_DETAILS_FROM_SREG = True

        # wiki: https://opennode.atlassian.net/wiki/display/WD/AuthOpenID+plugin+configuration
        WALDUR_AUTH_OPENID = {
            # on user registration following name will be used for user's registration_method field
            'NAME': 'openid',
            'LOGOUT_URL': 'https://openid.ee/auth/logout',
            'COUNTRY_CODE': 'EE',
        }

        OPENID_RENDER_FAILURE = login_failed

    @staticmethod
    def get_public_settings():
        return ['LOGOUT_URL']

    @staticmethod
    def update_settings(settings):
        # Enable customer OpenID authentication backend
        settings['AUTHENTICATION_BACKENDS'] += ('waldur_auth_openid.auth.WaldurOpenIDBackend',)

        # Enable app in order to connect database models and migrations
        settings['INSTALLED_APPS'] += ('django_openid_auth',)

    @staticmethod
    def django_app():
        return 'waldur_auth_openid'

    @staticmethod
    def django_urls():
        from .urls import urlpatterns
        return urlpatterns



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations

import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0076_plan_period_uuid_populate'),
    ]

    operations = [
        migrations.AlterField(
            model_name='resourceplanperiod',
            name='uuid',
            field=waldur_core.core.fields.UUIDField(),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-09-03 10:05
from __future__ import unicode_literals

from decimal import Decimal
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0030_add_required_field_to_attribute_model'),
    ]

    operations = [
        migrations.CreateModel(
            name='PlanComponent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.CharField(max_length=50)),
                ('amount', models.PositiveIntegerField(default=0)),
                ('price', models.DecimalField(decimal_places=10, default=0, max_digits=14, validators=[django.core.validators.MinValueValidator(Decimal('0'))], verbose_name='Price per unit per billing period.')),
                ('plan', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='components', to='marketplace.Plan')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='plancomponent',
            unique_together=set([('type', 'plan')]),
        ),
    ]



from __future__ import unicode_literals

import logging

from django.core import exceptions as django_exceptions
from django.db import transaction, IntegrityError

from waldur_core.core.models import StateMixin
from waldur_core.quotas.models import Quota
from waldur_core.structure import models as structure_models

from ..openstack import models as openstack_models, apps as openstack_apps
from . import apps, log, models, utils


logger = logging.getLogger(__name__)


def _log_scheduled_action(resource, action, action_details):
    class_name = resource.__class__.__name__.lower()
    message = _get_action_message(action, action_details)
    log.event_logger.openstack_resource_action.info(
        'Operation "%s" has been scheduled for %s "%s"' % (message, class_name, resource.name),
        event_type=_get_action_event_type(action, 'scheduled'),
        event_context={'resource': resource, 'action_details': action_details},
    )


def _log_succeeded_action(resource, action, action_details):
    if not action:
        return
    class_name = resource.__class__.__name__.lower()
    message = _get_action_message(action, action_details)
    log.event_logger.openstack_resource_action.info(
        'Successfully executed "%s" operation for %s "%s"' % (message, class_name, resource.name),
        event_type=_get_action_event_type(action, 'succeeded'),
        event_context={'resource': resource, 'action_details': action_details},
    )


def _log_failed_action(resource, action, action_details):
    class_name = resource.__class__.__name__.lower()
    message = _get_action_message(action, action_details)
    log.event_logger.openstack_resource_action.warning(
        'Failed to execute "%s" operation for %s "%s"' % (message, class_name, resource.name),
        event_type=_get_action_event_type(action, 'failed'),
        event_context={'resource': resource, 'action_details': action_details},
    )


def _get_action_message(action, action_details):
    return action_details.pop('message', action)


def _get_action_event_type(action, event_state):
    return 'resource_%s_%s' % (action.replace(' ', '_').lower(), event_state)


def log_action(sender, instance, created=False, **kwargs):
    """ Log any resource action.

        Example of logged volume extend action:
        {
            'event_type': 'volume_extend_succeeded',
            'message': 'Successfully executed "Extend volume from 1024 MB to 2048 MB" operation for volume "pavel-test"',
            'action_details': {'old_size': 1024, 'new_size': 2048}
        }
    """
    resource = instance
    if created or not resource.tracker.has_changed('action'):
        return
    if resource.state == StateMixin.States.UPDATE_SCHEDULED:
        _log_scheduled_action(resource, resource.action, resource.action_details)
    if resource.state == StateMixin.States.OK:
        _log_succeeded_action(
            resource, resource.tracker.previous('action'), resource.tracker.previous('action_details'))
    elif resource.state == StateMixin.States.ERRED:
        _log_failed_action(
            resource, resource.tracker.previous('action'), resource.tracker.previous('action_details'))


def log_snapshot_schedule_creation(sender, instance, created=False, **kwargs):
    if not created:
        return

    snapshot_schedule = instance
    log.event_logger.openstack_snapshot_schedule.info(
        'Snapshot schedule "%s" has been created' % snapshot_schedule.name,
        event_type='resource_snapshot_schedule_created',
        event_context={'resource': snapshot_schedule.source_volume, 'snapshot_schedule': snapshot_schedule},
    )


def log_snapshot_schedule_action(sender, instance, created=False, **kwargs):
    snapshot_schedule = instance
    if created or not snapshot_schedule.tracker.has_changed('is_active'):
        return

    context = {'resource': snapshot_schedule.source_volume, 'snapshot_schedule': snapshot_schedule}
    if snapshot_schedule.is_active:
        log.event_logger.openstack_snapshot_schedule.info(
            'Snapshot schedule "%s" has been activated' % snapshot_schedule.name,
            event_type='resource_snapshot_schedule_activated',
            event_context=context,
        )
    else:
        if snapshot_schedule.error_message:
            message = 'Snapshot schedule "%s" has been deactivated because of error: %s' % (
                snapshot_schedule.name, snapshot_schedule.error_message)
        else:
            message = 'Snapshot schedule "%s" has been deactivated' % snapshot_schedule.name
        log.event_logger.openstack_snapshot_schedule.warning(
            message,
            event_type='resource_snapshot_schedule_deactivated',
            event_context=context,
        )


def log_snapshot_schedule_deletion(sender, instance, **kwargs):
    snapshot_schedule = instance
    log.event_logger.openstack_snapshot_schedule.info(
        'Snapshot schedule "%s" has been deleted' % snapshot_schedule.name,
        event_type='resource_snapshot_schedule_deleted',
        event_context={'resource': snapshot_schedule.source_volume, 'snapshot_schedule': snapshot_schedule},
    )


def log_backup_schedule_creation(sender, instance, created=False, **kwargs):
    if not created:
        return

    backup_schedule = instance
    log.event_logger.openstack_backup_schedule.info(
        'Backup schedule "%s" has been created' % backup_schedule.name,
        event_type='resource_backup_schedule_created',
        event_context={'resource': backup_schedule.instance, 'backup_schedule': backup_schedule},
    )


def log_backup_schedule_action(sender, instance, created=False, **kwargs):
    backup_schedule = instance
    if created or not backup_schedule.tracker.has_changed('is_active'):
        return

    context = {'resource': backup_schedule.instance, 'backup_schedule': backup_schedule}
    if backup_schedule.is_active:
        log.event_logger.openstack_backup_schedule.info(
            'Backup schedule "%s" has been activated' % backup_schedule.name,
            event_type='resource_backup_schedule_activated',
            event_context=context,
        )
    else:
        if backup_schedule.error_message:
            message = 'Backup schedule "%s" has been deactivated because of error: %s' % (
                backup_schedule.name, backup_schedule.error_message)
        else:
            message = 'Backup schedule "%s" has been deactivated' % backup_schedule.name
        log.event_logger.openstack_backup_schedule.warning(
            message,
            event_type='resource_backup_schedule_deactivated',
            event_context=context,
        )


def log_backup_schedule_deletion(sender, instance, **kwargs):
    backup_schedule = instance
    log.event_logger.openstack_backup_schedule.info(
        'Backup schedule "%s" has been deleted' % backup_schedule.name,
        event_type='resource_backup_schedule_deleted',
        event_context={'resource': backup_schedule.instance, 'backup_schedule': backup_schedule},
    )


def update_service_settings_credentials(sender, instance, created=False, **kwargs):
    """
    Updates service settings credentials on tenant user_password or user_username change.
    It is possible to change a user password in tenant,
    as service settings copies tenant user password on creation it has to be update on change.
    """
    if created:
        return

    tenant = instance
    if tenant.tracker.has_changed('user_password') or tenant.tracker.has_changed('user_username'):
        service_settings = structure_models.ServiceSettings.objects.filter(scope=tenant).first()
        if service_settings:
            service_settings.username = tenant.user_username
            service_settings.password = tenant.user_password
            service_settings.save()


class BaseSynchronizationHandler(object):
    """
    This class provides signal handlers for synchronization of OpenStack properties
    when parent OpenStack resource are created, updated or deleted.
    Security groups, floating IPs, networks and subnets are implemented as
    resources in openstack application. However they are implemented as service properties
    in the openstack_tenant application.
    """
    property_model = None
    resource_model = None
    fields = []

    def get_tenant(self, resource):
        return resource.tenant

    def get_service_settings(self, resource):
        try:
            return structure_models.ServiceSettings.objects.get(scope=self.get_tenant(resource),
                                                                type=apps.OpenStackTenantConfig.service_name)
        except (django_exceptions.ObjectDoesNotExist, django_exceptions.MultipleObjectsReturned):
            return

    def get_service_property(self, resource, settings):
        try:
            return self.property_model.objects.get(settings=settings, backend_id=resource.backend_id)
        except (django_exceptions.ObjectDoesNotExist, django_exceptions.MultipleObjectsReturned):
            return

    def map_resource_to_dict(self, resource):
        return {field: getattr(resource, field) for field in self.fields}

    def create_service_property(self, resource, settings):
        defaults = dict(name=resource.name, **self.map_resource_to_dict(resource))

        try:
            with transaction.atomic():
                return self.property_model.objects.get_or_create(
                    settings=settings,
                    backend_id=resource.backend_id,
                    defaults=defaults
                )
        except IntegrityError:
            logger.warning('Could not create %s with backend ID %s '
                           'and service settings %s due to concurrent update.',
                           self.property_model, resource.backend_id, settings)

    def update_service_property(self, resource, settings):
        service_property = self.get_service_property(resource, settings)
        if not service_property:
            return
        params = self.map_resource_to_dict(resource)
        for key, value in params.items():
            setattr(service_property, key, value)
        service_property.name = resource.name
        service_property.save()
        return service_property

    def create_handler(self, sender, instance, name, source, target, **kwargs):
        """
        Creates service property on resource transition from 'CREATING' state to 'OK'.
        """
        if source == StateMixin.States.CREATING and target == StateMixin.States.OK:
            settings = self.get_service_settings(instance)
            if settings and not self.get_service_property(instance, settings):
                self.create_service_property(instance, settings)

    def import_handler(self, sender, instance, created=False, **kwargs):
        """
        Creates service property on when resource is imported.
        """
        if created and instance.state == StateMixin.States.OK:
            settings = self.get_service_settings(instance)
            if settings and not self.get_service_property(instance, settings):
                self.create_service_property(instance, settings)

    def update_handler(self, sender, instance, name, source, target, **kwargs):
        """
        Updates service property on resource transition from 'UPDATING' state to 'OK'.
        """
        if source == StateMixin.States.UPDATING and target == StateMixin.States.OK:
            settings = self.get_service_settings(instance)
            if settings:
                self.update_service_property(instance, settings)

    def delete_handler(self, sender, instance, **kwargs):
        """
        Deletes service property on resource deletion
        """
        settings = self.get_service_settings(instance)
        if not settings:
            return
        service_property = self.get_service_property(instance, settings)
        if not service_property:
            return
        service_property.delete()


class FloatingIPHandler(BaseSynchronizationHandler):
    property_model = models.FloatingIP
    resource_model = openstack_models.FloatingIP
    fields = ('address', 'backend_network_id', 'runtime_state')


class SecurityGroupHandler(BaseSynchronizationHandler):
    property_model = models.SecurityGroup
    resource_model = openstack_models.SecurityGroup
    fields = ('description',)

    def map_rules(self, security_group, openstack_security_group):
        return [models.SecurityGroupRule(
            protocol=rule.protocol,
            from_port=rule.from_port,
            to_port=rule.to_port,
            cidr=rule.cidr,
            backend_id=rule.backend_id,
            security_group=security_group,
        ) for rule in openstack_security_group.rules.iterator()]

    def create_service_property(self, resource, settings):
        service_property, _ = super(SecurityGroupHandler, self).create_service_property(resource, settings)
        if resource.rules.count() > 0:
            group_rules = self.map_rules(service_property, resource)
            service_property.rules.bulk_create(group_rules)
        return service_property

    def update_service_property(self, resource, settings):
        service_property = super(SecurityGroupHandler, self).update_service_property(resource, settings)
        if not service_property:
            return

        service_property.rules.all().delete()
        group_rules = self.map_rules(service_property, resource)
        service_property.rules.bulk_create(group_rules)
        return service_property


class NetworkHandler(BaseSynchronizationHandler):
    property_model = models.Network
    resource_model = openstack_models.Network
    fields = ('is_external', 'segmentation_id', 'type')


class SubNetHandler(BaseSynchronizationHandler):
    property_model = models.SubNet
    resource_model = openstack_models.SubNet
    fields = ('allocation_pools', 'cidr', 'dns_nameservers', 'enable_dhcp', 'ip_version')

    def get_tenant(self, resource):
        return resource.network.tenant

    def map_resource_to_dict(self, resource):
        params = super(SubNetHandler, self).map_resource_to_dict(resource)
        params['network'] = models.Network.objects.get(backend_id=resource.network.backend_id)
        return params


resource_handlers = (
    FloatingIPHandler(),
    SecurityGroupHandler(),
    NetworkHandler(),
    SubNetHandler(),
)


def sync_certificates_between_openstack_service_with_openstacktenant_service(sender, instance, action, **kwargs):
    """
    Copies certifications links in original service settings to derived openstack tenant service settings.
    Handling works only for OpenStack service settings and ignored for all others.
    """
    service_settings = instance
    if (action not in ['post_add', 'post_remove', 'post_clear'] or
            service_settings.type != openstack_apps.OpenStackConfig.service_name):
        return

    tenants = openstack_models.Tenant.objects.filter(service_project_link__service__settings=service_settings)

    if not tenants:
        return

    openstack_settings = structure_models.ServiceSettings.objects.filter(scope__in=tenants)

    with transaction.atomic():
        for settings in openstack_settings:
            settings.certifications.clear()
            settings.certifications.add(*service_settings.certifications.all())


def copy_certifications_from_openstack_service_to_openstacktenant_service(sender, instance, created=False, **kwargs):
    if not created or instance.type != apps.OpenStackTenantConfig.service_name:
        return

    tenant = instance.scope
    if not isinstance(tenant, openstack_models.Tenant):
        return

    admin_settings = tenant.service_project_link.service.settings

    with transaction.atomic():
        instance.certifications.clear()
        instance.certifications.add(*admin_settings.certifications.all())


def copy_flavor_exclude_regex_to_openstacktenant_service_settings(sender, instance, created=False, **kwargs):
    if not created or instance.type != apps.OpenStackTenantConfig.service_name:
        return

    tenant = instance.scope
    if not isinstance(tenant, openstack_models.Tenant):
        return

    admin_settings = tenant.service_project_link.service.settings
    instance.options['flavor_exclude_regex'] = admin_settings.options.get('flavor_exclude_regex', '')
    instance.save(update_fields=['options'])


def create_service_from_tenant(sender, instance, created=False, **kwargs):
    if not created:
        return

    if structure_models.ServiceSettings.objects.filter(
        scope=instance,
        type=apps.OpenStackTenantConfig.service_name,
    ).exists():
        return

    tenant = instance
    admin_settings = tenant.service_project_link.service.settings
    customer = tenant.service_project_link.project.customer
    service_settings = structure_models.ServiceSettings.objects.create(
        name=tenant.name,
        scope=tenant,
        customer=customer,
        type=apps.OpenStackTenantConfig.service_name,
        backend_url=admin_settings.backend_url,
        username=tenant.user_username,
        password=tenant.user_password,
        domain=admin_settings.domain,
        options={
            'availability_zone': tenant.availability_zone,
            'tenant_id': tenant.backend_id,
        },
    )

    if admin_settings.options.get('console_type'):
        service_settings.options['console_type'] = admin_settings.options.get('console_type')
        service_settings.save()

    service = models.OpenStackTenantService.objects.create(
        settings=service_settings,
        customer=customer,
    )

    models.OpenStackTenantServiceProjectLink.objects.create(
        service=service,
        project=tenant.service_project_link.project,
    )


def update_service_settings(sender, instance, created=False, **kwargs):
    tenant = instance

    if created or not (set(['external_network_id', 'name']) & set(tenant.tracker.changed())):
        return

    try:
        service_settings = structure_models.ServiceSettings.objects.get(scope=tenant,
                                                                        type=apps.OpenStackTenantConfig.service_name)
    except structure_models.ServiceSettings.DoesNotExist:
        return
    else:
        service_settings.options['external_network_id'] = tenant.external_network_id
        service_settings.name = tenant.name
        service_settings.save()


def sync_price_list_item_for_flavor(sender, instance, created=False, **kwargs):
    if created:
        utils.sync_price_list_item(instance)


def sync_private_settings_quotas_with_tenant_quotas(sender, instance, created=False, **kwargs):
    if created:
        return

    quota = instance
    if not isinstance(quota.scope, openstack_models.Tenant):
        return

    tenant = quota.scope
    private_settings = structure_models.ServiceSettings.objects.filter(scope=tenant)

    Quota.objects\
        .filter(scope__in=private_settings, name=quota.name)\
        .update(limit=quota.limit, usage=quota.usage)



from __future__ import unicode_literals

import django_filters

from waldur_core.quotas import models


class QuotaFilter(django_filters.NumberFilter):
    """
    Filter by quota value
    """

    def __init__(self, quota_name, quota_field, **kwargs):
        super(QuotaFilter, self).__init__(**kwargs)
        self.quota_name = quota_name
        self.quota_field = quota_field

    def filter(self, qs, value):
        return qs.filter(**{'quotas__name': self.quota_name, 'quotas__{}'.format(self.quota_field): value})


class QuotaFilterSet(django_filters.FilterSet):
    """
    FilterSet for quotas view
    """
    name = django_filters.CharFilter(
        lookup_expr='icontains',
    )

    class Meta(object):
        model = models.Quota
        fields = [
            'name'
        ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-06-07 10:02
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('packages', '0011_add_openstack_component_details_to_tenant'),
    ]

    operations = [
        migrations.AddField(
            model_name='packagetemplate',
            name='product_code',
            field=models.CharField(blank=True, max_length=30),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-05-07 13:42
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('packages', '0013_packagetemplate_article_code'),
    ]

    operations = [
        migrations.AlterField(
            model_name='openstackpackage',
            name='tenant',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='openstack.Tenant'),
        ),
    ]



from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers

from waldur_core.core import serializers as core_serializers

from . import models, utils


class ProfileSerializer(core_serializers.AugmentedSerializerMixin,
                        serializers.HyperlinkedModelSerializer):

    agree_with_policy = serializers.BooleanField(write_only=True,
                                                 help_text=_('User must agree with the policy.'))

    class Meta(object):
        model = models.Profile
        fields = ('uuid', 'username', 'agreement_date', 'is_active', 'agree_with_policy')
        protected_fields = ('username', 'agreement_date')
        read_only_fields = ('is_active',)

        extra_kwargs = dict(
            url={'lookup_field': 'uuid', 'view_name': 'freeipa-profile-detail'},
        )

    def create(self, validated_data):
        # Check if user already has FreeIPA profile
        user = self.context['request'].user
        if models.Profile.objects.filter(user=user).exists():
            raise serializers.ValidationError({
                'details': _('User already has profile.')
            })
        validated_data['user'] = user

        # Check if user agrees with policy
        if not validated_data.pop('agree_with_policy'):
            raise serializers.ValidationError({
                'agree_with_policy': _('User must agree with the policy.')
            })

        validated_data['username'] = utils.generate_username(validated_data['username'])

        return super(ProfileSerializer, self).create(validated_data)



from __future__ import unicode_literals

from django.db import transaction
from django.db.models import Q
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets, views, permissions, decorators, response, status, exceptions as rf_exceptions

from waldur_core.core import validators as core_validators
from waldur_core.core import views as core_views
from waldur_core.core import utils as core_utils
from waldur_core.structure import filters as structure_filters
from waldur_core.structure import metadata as structure_metadata
from waldur_core.structure import models as structure_models
from waldur_core.structure import permissions as structure_permissions
from waldur_core.structure import views as structure_views

from . import filters, models, serializers, backend, exceptions, tasks


class CheckExtensionMixin(core_views.CheckExtensionMixin):
    extension_name = 'WALDUR_SUPPORT'


class IssueViewSet(CheckExtensionMixin, core_views.ActionsViewSet):
    queryset = models.Issue.objects.all()
    lookup_field = 'uuid'
    filter_backends = (
        filters.IssueCallerOrRoleFilterBackend,
        DjangoFilterBackend,
        filters.IssueResourceFilterBackend,
    )
    filter_class = filters.IssueFilter
    serializer_class = serializers.IssueSerializer

    def is_staff_or_support(request, view, obj=None):
        if not request.user.is_staff and not request.user.is_support:
            raise rf_exceptions.PermissionDenied()

    def check_related_resources(request, view, obj=None):
        if obj and obj.offering_set.exists():
            raise rf_exceptions.ValidationError(_('Issue has offering. Please remove it first.'))

    def can_create_user(request, view, obj=None):
        if not request.user.email:
            raise rf_exceptions.ValidationError(_('Current user does not have email, '
                                                  'therefore he is not allowed to create issues.'))

        if not request.user.full_name:
            raise rf_exceptions.ValidationError(_('Current user does not have full_name, '
                                                  'therefore he is not allowed to create issues.'))

    @transaction.atomic()
    def perform_create(self, serializer):
        issue = serializer.save()
        try:
            backend.get_active_backend().create_issue(issue)
        except exceptions.SupportUserInactive:
            raise rf_exceptions.ValidationError({'caller': _('Caller is inactive.')})

    create_permissions = [can_create_user]

    @transaction.atomic()
    def perform_update(self, serializer):
        issue = serializer.save()
        backend.get_active_backend().update_issue(issue)

    update_permissions = partial_update_permissions = [is_staff_or_support]

    @transaction.atomic()
    def perform_destroy(self, issue):
        backend.get_active_backend().delete_issue(issue)
        issue.delete()

    destroy_permissions = [is_staff_or_support, check_related_resources]

    def _comment_permission(request, view, obj=None):
        user = request.user
        if user.is_staff or user.is_support or not obj:
            return
        issue = obj
        if issue.customer and issue.customer.has_user(user, structure_models.CustomerRole.OWNER):
            return
        if (issue.project and (issue.project.has_user(user, structure_models.ProjectRole.ADMINISTRATOR) or
                               issue.project.has_user(user, structure_models.ProjectRole.MANAGER))):
            return
        raise rf_exceptions.PermissionDenied()

    @decorators.detail_route(methods=['post'])
    def comment(self, request, uuid=None):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        with transaction.atomic():
            comment = serializer.save()
            backend.get_active_backend().create_comment(comment)
        return response.Response(serializer.data, status=status.HTTP_201_CREATED)

    comment_serializer_class = serializers.CommentSerializer
    comment_permissions = [_comment_permission]


class CommentViewSet(CheckExtensionMixin, core_views.ActionsViewSet):
    lookup_field = 'uuid'
    serializer_class = serializers.CommentSerializer
    filter_backends = (
        filters.CommentIssueCallerOrRoleFilterBackend,
        DjangoFilterBackend,
        filters.CommentIssueResourceFilterBackend,
    )
    filter_class = filters.CommentFilter
    queryset = models.Comment.objects.all()

    @transaction.atomic()
    def perform_update(self, serializer):
        comment = serializer.save()
        backend.get_active_backend().update_comment(comment)

    update_permissions = partial_update_permissions = [structure_permissions.is_staff]

    @transaction.atomic()
    def perform_destroy(self, comment):
        backend.get_active_backend().delete_comment(comment)
        comment.delete()

    destroy_permissions = [structure_permissions.is_staff]

    def get_queryset(self):
        queryset = super(CommentViewSet, self).get_queryset()

        if not self.request.user.is_staff:
            subquery = Q(is_public=True) | Q(author__user=self.request.user)
            queryset = queryset.filter(subquery)

        return queryset


class IsStaffOrSupportUser(permissions.BasePermission):
    """
    Allows access only to staff or global support users.
    """

    def has_permission(self, request, view):
        return request.user.is_staff or request.user.is_support


class SupportUserViewSet(CheckExtensionMixin, viewsets.ReadOnlyModelViewSet):
    queryset = models.SupportUser.objects.all()
    lookup_field = 'uuid'
    permission_classes = (permissions.IsAuthenticated, IsStaffOrSupportUser,)
    serializer_class = serializers.SupportUserSerializer
    filter_backends = (DjangoFilterBackend,)
    filter_class = filters.SupportUserFilter


class WebHookReceiverView(CheckExtensionMixin, views.APIView):
    authentication_classes = ()
    permission_classes = ()
    serializer_class = serializers.WebHookReceiverSerializer

    def post(self, request):
        serializer = self.serializer_class(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return response.Response(status=status.HTTP_200_OK)


class OfferingViewSet(CheckExtensionMixin, core_views.ActionsViewSet):
    queryset = models.Offering.objects.all()
    serializer_class = serializers.OfferingSerializer
    lookup_field = 'uuid'
    metadata_class = structure_metadata.ActionsMetadata
    filter_backends = (
        structure_filters.GenericRoleFilter,
        DjangoFilterBackend,
    )
    filter_class = filters.OfferingFilter

    @decorators.list_route()
    def configured(self, request):
        summary_config = {}
        for template in models.OfferingTemplate.objects.all():
            summary_config[template.name] = template.config
        return response.Response(summary_config, status=status.HTTP_200_OK)

    @transaction.atomic()
    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        offering = serializer.save()
        serialized_issue = core_utils.serialize_instance(offering.issue)
        task = tasks.create_issue.s(serialized_issue)
        transaction.on_commit(lambda: task.apply_async(countdown=2))
        return response.Response(serializer.data, status=status.HTTP_201_CREATED)

    create_serializer_class = serializers.OfferingCreateSerializer
    create_permissions = [structure_permissions.is_owner,
                          structure_permissions.is_manager,
                          structure_permissions.is_administrator]

    def offering_is_in_requested_state(offering):
        if offering.state != models.Offering.States.REQUESTED:
            raise rf_exceptions.ValidationError(_('Offering must be in requested state.'))

    @decorators.detail_route(methods=['post'])
    def complete(self, request, uuid=None):
        serializer = self.get_serializer(instance=self.get_object(), data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return response.Response({'status': _('Offering is marked as completed.')}, status=status.HTTP_200_OK)

    complete_validators = [offering_is_in_requested_state]
    complete_permissions = [structure_permissions.is_staff]
    complete_serializer_class = serializers.OfferingCompleteSerializer

    @decorators.detail_route(methods=['post'])
    def terminate(self, request, uuid=None):
        offering = self.get_object()
        offering.state = models.Offering.States.TERMINATED
        offering.terminated_at = timezone.now()
        offering.save()
        return response.Response({'status': _('Offering is marked as terminated.')}, status=status.HTTP_200_OK)

    terminate_permissions = [structure_permissions.is_staff]

    update_permissions = partial_update_permissions = [structure_permissions.is_staff]

    destroy_permissions = [structure_permissions.is_staff]
    destroy_validators = [core_validators.StateValidator(models.Offering.States.TERMINATED)]


class AttachmentViewSet(CheckExtensionMixin,
                        core_views.ActionsViewSet):
    queryset = models.Attachment.objects.all()
    filter_class = filters.AttachmentFilter
    filter_backends = [DjangoFilterBackend]
    serializer_class = serializers.AttachmentSerializer
    lookup_field = 'uuid'
    disabled_actions = ['update', 'partial_update']

    @transaction.atomic()
    def perform_destroy(self, attachment):
        backend.get_active_backend().delete_attachment(attachment)
        attachment.delete()

    @transaction.atomic()
    def perform_create(self, serializer):
        attachment = serializer.save()
        backend.get_active_backend().create_attachment(attachment)

    def get_queryset(self):
        queryset = super(AttachmentViewSet, self).get_queryset()

        if not self.request.user.is_staff:
            user_customers = structure_models.Customer.objects.filter(
                permissions__role=structure_models.CustomerRole.OWNER,
                permissions__user=self.request.user,
                permissions__is_active=True)
            subquery = Q(issue__customer__in=user_customers) | Q(issue__caller=self.request.user)
            queryset = queryset.filter(subquery)

        return queryset


class TemplateViewSet(CheckExtensionMixin, viewsets.ReadOnlyModelViewSet):
    permission_classes = (permissions.IsAuthenticated,)
    queryset = models.Template.objects.all()
    lookup_field = 'uuid'
    serializer_class = serializers.TemplateSerializer


class OfferingTemplateViewSet(viewsets.ReadOnlyModelViewSet):
    permission_classes = (permissions.IsAuthenticated,)
    queryset = models.OfferingTemplate.objects.all()
    lookup_field = 'uuid'
    serializer_class = serializers.OfferingTemplateSerializer


class OfferingPlanViewSet(viewsets.ReadOnlyModelViewSet):
    permission_classes = (permissions.IsAuthenticated,)
    queryset = models.OfferingPlan.objects.all()
    lookup_field = 'uuid'
    serializer_class = serializers.OfferingPlanSerializer


def get_offerings_count(scope):
    return scope.quotas.get(name='nc_offering_count').usage


structure_views.CustomerCountersView.register_counter('offerings', get_offerings_count)
structure_views.ProjectCountersView.register_counter('offerings', get_offerings_count)



from __future__ import unicode_literals

from django.db import models
from model_utils import FieldTracker

from waldur_azure import validators
from waldur_core.core import models as core_models
from waldur_core.quotas.fields import CounterQuotaField
from waldur_core.quotas.models import QuotaModelMixin
from waldur_core.structure import models as structure_models


class AzureService(structure_models.Service):
    projects = models.ManyToManyField(
        structure_models.Project, related_name='azure_services', through='AzureServiceProjectLink')

    @classmethod
    def get_url_name(cls):
        return 'azure'

    class Quotas(QuotaModelMixin.Quotas):
        vm_count = CounterQuotaField(
            target_models=lambda: [VirtualMachine],
            path_to_scope='service_project_link.service'
        )


class AzureServiceProjectLink(structure_models.ServiceProjectLink):
    service = models.ForeignKey(AzureService)

    @classmethod
    def get_url_name(cls):
        return 'azure-spl'


class Location(core_models.CoordinatesMixin,
               structure_models.ServiceProperty):

    enabled = models.BooleanField(
        default=True,
        help_text='Indicates whether location is available for resource group.'
    )

    class Meta:
        ordering = ['name']

    @classmethod
    def get_url_name(cls):
        return 'azure-location'


class Image(structure_models.ServiceProperty):
    publisher = models.CharField(max_length=255)
    sku = models.CharField(max_length=255)
    version = models.CharField(max_length=255)

    class Meta(structure_models.ServiceProperty.Meta):
        ordering = ['publisher', 'name', 'sku']

    @classmethod
    def get_url_name(cls):
        return 'azure-image'


class Size(structure_models.ServiceProperty):
    max_data_disk_count = models.PositiveIntegerField()
    memory_in_mb = models.PositiveIntegerField()
    number_of_cores = models.PositiveIntegerField()
    os_disk_size_in_mb = models.PositiveIntegerField()
    resource_disk_size_in_mb = models.PositiveIntegerField()

    class Meta(structure_models.ServiceProperty.Meta):
        ordering = ['number_of_cores', 'memory_in_mb']

    @classmethod
    def get_url_name(cls):
        return 'azure-size'


class BaseResource(core_models.RuntimeStateMixin, structure_models.NewResource):
    service_project_link = models.ForeignKey(AzureServiceProjectLink)

    class Meta(object):
        abstract = True


class ResourceGroup(BaseResource):
    name = models.CharField(max_length=90, validators=[validators.ResourceGroupNameValidator])
    location = models.ForeignKey(Location)

    @classmethod
    def get_url_name(cls):
        return 'azure-resource-group'


class BaseResourceGroupModel(BaseResource):
    resource_group = models.ForeignKey(ResourceGroup)

    class Meta(object):
        abstract = True


class StorageAccount(BaseResourceGroupModel):
    name = models.CharField(max_length=24, validators=[validators.StorageAccountNameValidator])


class Network(BaseResourceGroupModel):
    name = models.CharField(max_length=64, validators=[validators.NetworkingNameValidator])
    cidr = models.CharField(max_length=32)


class SubNet(BaseResourceGroupModel):
    name = models.CharField(max_length=80, validators=[validators.NetworkingNameValidator])
    network = models.ForeignKey(Network)
    cidr = models.CharField(max_length=32)


class SecurityGroup(BaseResourceGroupModel):
    name = models.CharField(max_length=80, validators=[validators.NetworkingNameValidator])


class NetworkInterface(BaseResourceGroupModel):
    name = models.CharField(max_length=80, validators=[validators.NetworkingNameValidator])
    subnet = models.ForeignKey(SubNet)
    config_name = models.CharField(max_length=255)
    public_ip = models.ForeignKey('PublicIP', on_delete=models.SET_NULL, null=True, blank=True)
    security_group = models.ForeignKey(SecurityGroup, on_delete=models.SET_NULL, null=True, blank=True)
    ip_address = models.GenericIPAddressField(null=True, blank=True, protocol='IPv4', default=None)
    tracker = FieldTracker()


class PublicIP(BaseResourceGroupModel):
    name = models.CharField(max_length=80, validators=[validators.NetworkingNameValidator])
    location = models.ForeignKey(Location)
    ip_address = models.GenericIPAddressField(null=True, blank=True, protocol='IPv4', default=None)
    tracker = FieldTracker()

    @classmethod
    def get_url_name(cls):
        return 'azure-public-ip'


class VirtualMachine(structure_models.VirtualMachine):
    service_project_link = models.ForeignKey(
        AzureServiceProjectLink, related_name='virtualmachines', on_delete=models.PROTECT)
    resource_group = models.ForeignKey(ResourceGroup)
    size = models.ForeignKey(Size)
    image = models.ForeignKey(Image)
    ssh_key = models.ForeignKey(core_models.SshPublicKey, null=True, blank=True)
    network_interface = models.ForeignKey(NetworkInterface)
    name = models.CharField(max_length=15, validators=[validators.VirtualMachineNameValidator])
    username = models.CharField(max_length=32, validators=[validators.VirtualMachineUsernameValidator])
    password = models.CharField(max_length=72, validators=validators.VirtualMachinePasswordValidators)
    user_data = models.TextField(
        blank=True,
        max_length=87380,
        help_text='Additional data that will be added to instance on provisioning')

    @property
    def internal_ips(self):
        return [self.network_interface.ip_address]

    @property
    def external_ips(self):
        public_ip = self.network_interface.public_ip
        return public_ip and [public_ip.ip_address] or []

    @classmethod
    def get_url_name(cls):
        return 'azure-virtualmachine'

    def get_access_url_name(self):
        return 'azure-virtualmachine-rdp'


class SQLServer(BaseResourceGroupModel):
    name = models.CharField(max_length=80, validators=[validators.SQLServerNameValidator])
    username = models.CharField(max_length=50, validators=[validators.SQLServerUsernameValidator])
    password = models.CharField(max_length=128, validators=validators.SQLServerPasswordValidators)
    storage_mb = models.PositiveIntegerField(null=True, validators=validators.SQLServerStorageValidators)
    fqdn = models.TextField(null=True, blank=True)
    tracker = FieldTracker()

    @classmethod
    def get_url_name(cls):
        return 'azure-sql-server'


class SQLDatabase(BaseResource):
    server = models.ForeignKey(SQLServer)
    charset = models.CharField(max_length=255, blank=True, null=True, default='utf8')
    collation = models.CharField(max_length=255, blank=True, null=True, default='utf8_general_ci')
    tracker = FieldTracker()

    @classmethod
    def get_url_name(cls):
        return 'azure-sql-database'



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-09-10 08:32
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0037_internal_ip_settings'),
    ]

    operations = [
        migrations.AlterField(
            model_name='internalip',
            name='settings',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


def migrate_items(apps, schema_editor):
    ContentType = apps.get_model('contenttypes', 'ContentType')
    GenericInvoiceItem = apps.get_model('invoices', 'GenericInvoiceItem')

    Offering = apps.get_model('support', 'Offering')
    OpenStackPackage = apps.get_model('packages', 'OpenStackPackage')

    offering_ct = ContentType.objects.get_for_model(Offering)
    package_ct = ContentType.objects.get_for_model(OpenStackPackage)

    for item in GenericInvoiceItem.objects.filter(object_id=None):
        if item.details.get('tenant_name') and item.details.get('tenant_uuid'):
            item.content_type = package_ct
            item.details.update({
                'name': item.details['tenant_name'],
                'scope_uuid': item.details['tenant_uuid'],
            })
            item.save()

        elif item.details.get('offering_name') and item.details.get('offering_uuid'):
            item.content_type = offering_ct
            item.details.update({
                'name': item.details['offering_name'],
                'scope_uuid': item.details['offering_uuid'],
            })
            item.save()


class Migration(migrations.Migration):

    dependencies = [
        # See also: https://code.djangoproject.com/ticket/24303
        ('contenttypes', '0002_remove_content_type_name'),
        ('invoices', '0028_delete_offering_and_openstack_item'),
    ]

    operations = [
        migrations.RunPython(migrate_items),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-08-22 20:29
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0023_add_billing'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='screenshots',
            new_name='screenshot',
        ),
    ]



from django.contrib import admin
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from waldur_core.core import admin as core_admin
from . import models, tasks


class IdentityProviderAdmin(core_admin.ExtraActionsMixin, admin.ModelAdmin):
    fields = ('name', 'url')
    readonly_fields = ('name', 'url')
    list_display = ('name',)
    search_fields = ('name',)

    def get_extra_actions(self):
        return [
            self.sync_providers,
        ]

    def sync_providers(self, request):
        tasks.sync_providers()
        self.message_user(request, _('Identity providers have been synchronized.'))
        return redirect(reverse('admin:waldur_auth_saml2_identityprovider_changelist'))


admin.site.register(models.IdentityProvider, IdentityProviderAdmin)



from django.core.management.base import CommandError

from waldur_core.core.utils import DryRunCommand
from waldur_mastermind.marketplace.models import Category
from waldur_mastermind.marketplace_openstack import utils


class Command(DryRunCommand):
    help = """Import OpenStack tenant service settings as marketplace offerings."""

    def handle(self, dry_run, *args, **options):
        try:
            offerings_counter, plans_counter = utils.import_openstack_tenant_service_settings(dry_run)
            self.stdout.write(self.style.SUCCESS('%s offerings have been created.' % offerings_counter))
            self.stdout.write(self.style.SUCCESS('%s plans have been created.' % plans_counter))
        except Category.DoesNotExist:
            raise CommandError('Please ensure that WALDUR_MARKETPLACE_OPENSTACK.INSTANCE_CATEGORY_UUID '
                               'and WALDUR_MARKETPLACE_OPENSTACK.VOLUME_CATEGORY_UUID'
                               'setting has valid value.')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-09-10 08:03
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


def populate_internal_ip_settings(apps, schema_editor):
    InternalIP = apps.get_model('openstack_tenant', 'InternalIP')
    for internal_ip in InternalIP.objects.all():
        internal_ip.settings = internal_ip.subnet.settings
        internal_ip.save()


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0002_immutable_default_json'),
        ('openstack_tenant', '0036_instance_shared_ips'),
    ]

    operations = [
        migrations.AddField(
            model_name='internalip',
            name='settings',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings'),
        ),
        migrations.AlterField(
            model_name='internalip',
            name='backend_id',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterUniqueTogether(
            name='internalip',
            unique_together=set([('backend_id', 'settings')]),
        ),
        migrations.RunPython(populate_internal_ip_settings),
    ]



import logging
from datetime import timedelta

from django.conf import settings
from django.utils import timezone
from influxdb import InfluxDBClient, exceptions
from reversion.models import Version

from waldur_core.structure.models import Customer, Project

from . import models


logger = logging.getLogger(__name__)


def get_influxdb_client():
    if settings.WALDUR_ANALYTICS['ENABLED']:
        options = settings.WALDUR_ANALYTICS['INFLUXDB']
        return InfluxDBClient(**options)


def write_points(client, points):
    try:
        client.write_points(points)
    except (exceptions.InfluxDBClientError, exceptions.InfluxDBServerError) as e:
        logger.warning('Unable to write to InfluxDB %s', e)


def import_daily_usage():
    quotas = {}
    cutoff = timezone.now() - timedelta(days=90)
    versions = Version.objects.filter(
        revision__date_created__gte=cutoff
    ).order_by('revision__date_created')
    for version in versions:
        try:
            scope = version.object.scope
        except AttributeError:
            continue
        if not isinstance(scope, (Customer, Project)):
            continue
        name = version.object.name
        usage = version._object_version.object.usage
        date = version.revision.date_created.date()

        quotas.setdefault(scope, {})
        quotas[scope].setdefault(name, {})
        quotas[scope][name][date] = usage

    end = timezone.now().date()
    for scope in quotas.keys():
        for name in quotas[scope].keys():
            records = quotas[scope][name]
            start = min(records.keys())
            days = (end - start).days
            usage = 0
            for i in range(days + 1):
                date = start + timedelta(days=i)
                usage = records.get(date, usage)
                models.DailyQuotaHistory.objects.update_or_create_quota(scope, name, date, usage)



from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.db import models as django_models

from waldur_core.core import managers as core_managers
from waldur_core.structure import models as structure_models


class MixinManager(core_managers.GenericKeyMixin, django_models.Manager):
    pass


class OfferingQuerySet(django_models.QuerySet):
    def filter_for_user(self, user):
        if user.is_staff or user.is_support:
            return self

        connected_customers = set(structure_models.Customer.objects.all().filter(
            Q(permissions__user=user, permissions__is_active=True) |
            Q(projects__permissions__user=user, projects__permissions__is_active=True)
        ).distinct())

        return self.filter(
            Q(shared=True) |
            Q(shared=False, allowed_customers__in=connected_customers) |
            Q(shared=False, customer__in=connected_customers)
        )

    def filter_for_customer(self, value):
        return self.filter(Q(shared=True) |
                           Q(customer__uuid=value) |
                           Q(allowed_customers__uuid=value))

    def filter_for_project(self, value):
        settings_ct = ContentType.objects.get_for_model(structure_models.ServiceSettings)
        service_settings = {
            pk
            for spl in structure_models.ServiceProjectLink.get_all_models()
            for pk in spl.objects.filter(project__uuid=value).values_list('service__settings_id', flat=True)
        }

        return self.filter(
            Q(content_type=settings_ct, object_id__in=service_settings) |
            ~Q(content_type=settings_ct)
        )


class OfferingManager(MixinManager):
    def get_queryset(self):
        return OfferingQuerySet(self.model, using=self._db)



from __future__ import unicode_literals

from django.conf.urls import url

from . import views


def register_in(router):
    router.register(r'paypal-payments', views.PaymentView, base_name='paypal-payment')
    router.register(r'paypal-invoices', views.InvoicesViewSet, base_name='paypal-invoice')


urlpatterns = [
    url(r'^api/paypal-invoices-webhook/$', views.InvoiceWebHookViewSet.as_view(), name='paypal-invoice-webhook'),
]



import logging

from waldur_slurm.base import BatchError, BaseBatchClient
from waldur_slurm.parser_moab import MoabReportLine
from waldur_slurm.structures import Account, Association
from waldur_slurm.utils import format_current_month


class MoabError(BatchError):
    pass


logger = logging.getLogger(__name__)


class MoabClient(BaseBatchClient):
    """
    This class implements Python client for MOAB.
    See also MOAB Accounting Manager 9.1.1 Administrator Guide
    http://docs.adaptivecomputing.com/9-1-1/MAM/help.htm"""

    def list_accounts(self):
        output = self.execute_command(
            'mam-list-accounts --raw --quiet --show Name,Description,Organization'.split()
        )
        return [self._parse_account(line) for line in output.splitlines() if '|' in line]

    def _parse_account(self, line):
        parts = line.split('|')
        return Account(
            name=parts[0],
            description=parts[1],
            organization=parts[2],
        )

    def get_account(self, name):
        command = 'mam-list-accounts --raw --quiet --show Name,Description,Organization -a %s' % name
        output = self.execute_command(command.split())
        lines = [line for line in output.splitlines() if '|' in line]
        if len(lines) == 0:
            return None
        return self._parse_account(lines[0])

    def create_account(self, name, description, organization, parent_name=None):
        command = 'mam-create-account -a %(name)s -d "%(description)s" -o %(organization)s' % {
            'name': name,
            'description': description,
            'organization': organization,
        }
        return self.execute_command(command.split())

    def delete_account(self, name):
        command = 'mam-delete-account -a %s' % name
        return self.execute_command(command.split())

    def set_resource_limits(self, account, quotas):
        if quotas.deposit < 0:
            logger.warning('Skipping limit update because pricing '
                           'package is not created for the related service settings.')
            return
        command = 'mam-deposit -a %(account)s -z %(deposit_amount)s --create-fund True' % {
            'account': account,
            'deposit_amount': quotas.deposit
        }
        return self.execute_command(command.split())

    def get_association(self, user, account):
        command = 'mam-list-funds --raw --quiet -u %(user)s -a %(account)s --show Constraints,Balance' % \
                  {'user': user, 'account': account}
        output = self.execute_command(command.split())
        lines = [line for line in output.splitlines() if '|' in line]
        if len(lines) == 0:
            return None

        return Association(
            account=account,
            user=user,
            value=lines[0].split('|')[-1],
        )

    def create_association(self, username, account, default_account=None):
        command = 'mam-modify-account --add-user %(username)s -a %(account)s' % {
            'username': username,
            'account': account
        }
        return self.execute_command(command.split())

    def delete_association(self, username, account):
        command = 'mam-modify-account --del-user %(username)s -a %(account)s' % {
            'username': username,
            'account': account
        }
        return self.execute_command(command.split())

    def get_usage_report(self, accounts):
        template = (
            'mam-list-usagerecords --raw --quiet --show '
            'Account,Processors,GPUs,Memory,Duration,User,Charge,Nodes '
            '-a %(account)s -s %(start)s -e %(end)s'
        )
        month_start, month_end = format_current_month()

        report_lines = []
        for account in accounts:
            command = template % {
                'account': account,
                'start': month_start,
                'end': month_end,
            }
            lines = self.execute_command(command.split()).splitlines()
            for line in lines:
                if '|' in line:
                    report_lines.append(MoabReportLine(line))

        return report_lines



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2017-03-02 15:24
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


def connect_floating_ips_to_internal_ips(apps, schema_editor):
    Instance = apps.get_model('openstack_tenant', 'Instance')
    InternalIP = apps.get_model('openstack_tenant', 'InternalIP')

    for instance in Instance.objects.iterator():
        floating_ip = instance.external_ip
        internal_ip = InternalIP.objects.filter(instance=instance).first()
        # we assume that instance has only one internal IP.
        if not floating_ip or not internal_ip:
            continue
        floating_ip.internal_ip = internal_ip
        floating_ip.save()


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0021_instance_subnets'),
    ]

    operations = [
        migrations.AddField(
            model_name='floatingip',
            name='internal_ip',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='floating_ips', to='openstack_tenant.InternalIP'),
        ),
        migrations.RunPython(connect_floating_ips_to_internal_ips),
        migrations.AlterField(
            model_name='floatingip',
            name='address',
            field=models.GenericIPAddressField(null=True, protocol='IPv4'),
        ),
        migrations.AlterUniqueTogether(
            name='floatingip',
            unique_together=set([]),
        ),
        migrations.AddField(
            model_name='internalip',
            name='backend_id',
            field=models.CharField(blank=True, max_length=255),
        ),
    ]



from django.db import models as django_models


class SupportUserQuerySet(django_models.QuerySet):

    def get_or_create_from_user(self, user):
        """ Get or create support user based on regular user """
        return self.get_or_create(user=user, defaults={'name': user.full_name or user.username})


SupportUserManager = django_models.Manager.from_queryset(SupportUserQuerySet)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-03-05 21:05
from __future__ import unicode_literals

import django.utils.timezone
import django_fsm
import model_utils.fields
from django.conf import settings
from django.db import migrations, models

import waldur_core.core.fields
import waldur_core.core.validators


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ('openstack_tenant', '0030_add_volume_image_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='CachedRepositoryPythonLibrary',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('name', models.CharField(db_index=True, max_length=255, validators=[waldur_core.core.validators.validate_name])),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='InstalledLibrary',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('version', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PythonManagement',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('virtual_envs_dir_path', models.CharField(max_length=255)),
                ('python_version', models.CharField(max_length=10)),
                ('instance', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='openstack_tenant.Instance')),
                ('service_project_link', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='openstack_tenant.OpenStackTenantServiceProjectLink')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='PythonManagementDeleteRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(
                    choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'),
                             (4, 'Erred')], default=5)),
                ('virtual_env_name', models.CharField(max_length=255)),
                ('output', models.TextField(blank=True)),
                ('python_management', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='python_management.PythonManagement')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='PythonManagementDeleteVirtualEnvRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(
                    choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'),
                             (4, 'Erred')], default=5)),
                ('virtual_env_name', models.CharField(max_length=255)),
                ('output', models.TextField(blank=True)),
                ('python_management', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='python_management.PythonManagement')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='PythonManagementFindInstalledLibrariesRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(
                    choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'),
                             (4, 'Erred')], default=5)),
                ('virtual_env_name', models.CharField(max_length=255)),
                ('output', models.TextField(blank=True)),
                ('python_management', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='python_management.PythonManagement')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='PythonManagementFindVirtualEnvsRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(
                    choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'),
                             (4, 'Erred')], default=5)),
                ('virtual_env_name', models.CharField(max_length=255)),
                ('output', models.TextField(blank=True)),
                ('python_management', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='python_management.PythonManagement')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='PythonManagementInitializeRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(
                    choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'),
                             (4, 'Erred')], default=5)),
                ('output', models.TextField(blank=True)),
                ('python_management', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='python_management.PythonManagement')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='PythonManagementSynchronizeRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(
                    choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'),
                             (4, 'Erred')], default=5)),
                ('virtual_env_name', models.CharField(max_length=255)),
                ('output', models.TextField(blank=True)),
                ('libraries_to_install', waldur_core.core.fields.JSONField(blank=True, default=[], help_text='List of libraries to install')),
                ('libraries_to_remove', waldur_core.core.fields.JSONField(blank=True, default=[], help_text='List of libraries to remove')),
                ('initialization_request', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='sychronization_requests',
                                                             to='python_management.PythonManagementInitializeRequest')),
                ('python_management', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='python_management.PythonManagement')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='VirtualEnvironment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('python_management', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='virtual_environments', to='python_management.PythonManagement')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='installedlibrary',
            name='virtual_environment',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='installed_libraries', to='python_management.VirtualEnvironment'),
        ),
        migrations.AlterUniqueTogether(
            name='pythonmanagement',
            unique_together=set([('instance', 'virtual_envs_dir_path')]),
        ),
    ]



from django.utils.translation import ugettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend

from waldur_core.core import exceptions as core_exceptions
from waldur_core.core import mixins as core_mixins
from waldur_core.core import validators as core_validators
from waldur_core.core import views as core_views
from waldur_core.structure import filters as structure_filters
from waldur_core.structure import metadata as structure_metadata
from waldur_core.structure import models as structure_models
from waldur_core.structure import permissions as structure_permissions

from . import serializers, executors, filters, models


class PlaybookViewSet(core_views.ActionsViewSet):
    lookup_field = 'uuid'
    queryset = models.Playbook.objects.all().order_by('pk')
    unsafe_methods_permissions = [structure_permissions.is_staff]
    serializer_class = serializers.PlaybookSerializer


def check_all_related_resource_are_stable(job):
    States = structure_models.NewResource.States
    stable_states = (States.OK, States.ERRED)
    if not all(resource.state in stable_states for resource in job.get_related_resources()):
        raise core_exceptions.IncorrectStateException(_('Related resources are not stable yet. '
                                                        'Please wait until provisioning is completed.'))


class JobViewSet(core_mixins.CreateExecutorMixin, core_views.ActionsViewSet):
    lookup_field = 'uuid'
    queryset = models.Job.objects.all().order_by('pk')
    filter_backends = (structure_filters.GenericRoleFilter, DjangoFilterBackend)
    filter_class = filters.AnsibleJobsFilter
    unsafe_methods_permissions = [structure_permissions.is_administrator]
    serializer_class = serializers.JobSerializer
    metadata_class = structure_metadata.ActionsMetadata
    create_executor = executors.RunJobExecutor

    destroy_validators = [
        check_all_related_resource_are_stable,
        core_validators.StateValidator(models.Job.States.OK, models.Job.States.ERRED)
    ]
    delete_executor = executors.DeleteJobExecutor
